#!/usr/bin/env python3

import torch
from captum.attr._core.deep_lift import DeepLift, DeepLiftShap
from captum.attr._core.feature_ablation import FeatureAblation
from captum.attr._core.feature_permutation import FeaturePermutation
from captum.attr._core.gradient_shap import GradientShap
from captum.attr._core.guided_backprop_deconvnet import Deconvolution, GuidedBackprop
from captum.attr._core.guided_grad_cam import GuidedGradCam
from captum.attr._core.input_x_gradient import InputXGradient
from captum.attr._core.integrated_gradients import IntegratedGradients
from captum.attr._core.kernel_shap import KernelShap
from captum.attr._core.layer.grad_cam import LayerGradCam
from captum.attr._core.layer.internal_influence import InternalInfluence
from captum.attr._core.layer.layer_activation import LayerActivation
from captum.attr._core.layer.layer_conductance import LayerConductance
from captum.attr._core.layer.layer_deep_lift import LayerDeepLift, LayerDeepLiftShap
from captum.attr._core.layer.layer_feature_ablation import LayerFeatureAblation
from captum.attr._core.layer.layer_gradient_shap import LayerGradientShap
from captum.attr._core.layer.layer_gradient_x_activation import LayerGradientXActivation
from captum.attr._core.layer.layer_integrated_gradients import LayerIntegratedGradients
from captum.attr._core.layer.layer_lrp import LayerLRP
from captum.attr._core.lime import Lime
from captum.attr._core.lrp import LRP
from captum.attr._core.neuron.neuron_conductance import NeuronConductance
from captum.attr._core.neuron.neuron_deep_lift import NeuronDeepLift, NeuronDeepLiftShap
from captum.attr._core.neuron.neuron_feature_ablation import NeuronFeatureAblation
from captum.attr._core.neuron.neuron_gradient import NeuronGradient
from captum.attr._core.neuron.neuron_gradient_shap import NeuronGradientShap
from captum.attr._core.neuron.neuron_guided_backprop_deconvnet import (
    NeuronDeconvolution,
    NeuronGuidedBackprop,
)
from captum.attr._core.neuron.neuron_integrated_gradients import (
    NeuronIntegratedGradients,
)
from captum.attr._core.occlusion import Occlusion
from captum.attr._core.saliency import Saliency
from captum.attr._core.shapley_value import ShapleyValueSampling
from captum.attr._utils.input_layer_wrapper import ModelInputWrapper
from tests.helpers.basic import set_all_random_seeds
from tests.helpers.basic_models import (
    BasicModel_ConvNet,
    BasicModel_MultiLayer,
    BasicModel_MultiLayer_MultiInput,
    BasicModel_MultiLayer_TrueMultiInput,
    ReLULinearModel,
)

"""
This file defines a test configuration for attribution methods, particularly
defining valid input parameters for attribution methods. These test cases are
utilized for DataParallel tests, JIT tests, and target tests. Generally, these
tests follow a consistent structure of running the identified algorithm(s) in
two different way, e.g. with a DataParallel or JIT wrapped model versus a standard
model and verifying that the results match. New tests for additional model variants
or features can be built using this config.

The current schema for each test cases (each element in the list config) includes
the following information:
* "name": String defining name for test config
* "algorithms": List of algorithms (Attribution classes) which are applicable for
    the given test case
* "model": nn.Module model for given test
* "attribute_args": Arguments to be passed to attribute call of algorithm
* "layer": nn.Module corresponding to layer for Layer or Neuron attribution
* "noise_tunnel": True or False, based on whether to apply NoiseTunnel to the algorithm.
    If True, "attribute_args" corresponds to arguments for NoiseTunnel.attribute.
* "baseline_distr": True or False based on whether baselines in "attribute_args" are
    provided as a distribution or per-example.
* "target_delta": Delta for comparison in test_targets
* "dp_delta": Delta for comparison in test_data_parallel

To add tests for a new algorithm, simply add the algorithm to any existing test
case with applicable parameters by adding the algorithm to the corresponding
algorithms list. If the algorithm has particular arguments not covered by existing
test cases, add a new test case following the config schema described above. For
targets tests, ensure that the new test cases includes cases with tensor or list
targets. If the new algorithm works with JIT models, make sure to also
add the method to the whitelist in test_jit.

To create new tests for all methods, follow the same structure as test_jit,
test_targets, or test_data_parallel. Each of these iterates through the test
config and creates relevant test cases based on the config.
"""

# Set random seeds to ensure deterministic behavior
set_all_random_seeds(1234)

config = [
    # Attribution Method Configs
    # Primary Methods (Generic Configs)
    {
        "name": "basic_single_target",
        "algorithms": [
            IntegratedGradients,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            Saliency,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {"inputs": torch.randn(4, 3), "target": 1},
    },
    {
        "name": "basic_multi_input",
        "algorithms": [
            IntegratedGradients,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            Saliency,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "target": 0,
        },
        "dp_delta": 0.001,
    },
    {
        "name": "basic_multi_target",
        "algorithms": [
            IntegratedGradients,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            Saliency,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {"inputs": torch.randn(4, 3), "target": [0, 1, 1, 0]},
    },
    {
        "name": "basic_multi_input_multi_target",
        "algorithms": [
            IntegratedGradients,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            Saliency,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
        },
        "dp_delta": 0.0005,
    },
    {
        "name": "basic_multiple_tuple_target",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
        },
    },
    {
        "name": "basic_tensor_single_target",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {"inputs": torch.randn(4, 3), "target": torch.tensor([0])},
    },
    {
        "name": "basic_tensor_multi_target",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": torch.tensor([1, 1, 0, 0]),
        },
    },
    # Primary Configs with Baselines
    {
        "name": "basic_multiple_tuple_target_with_baselines",
        "algorithms": [
            IntegratedGradients,
            FeatureAblation,
            DeepLift,
            ShapleyValueSampling,
            Lime,
            KernelShap,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
        },
    },
    {
        "name": "basic_tensor_single_target_with_baselines",
        "algorithms": [
            IntegratedGradients,
            FeatureAblation,
            DeepLift,
            ShapleyValueSampling,
            Lime,
            KernelShap,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(4, 3),
            "target": torch.tensor([0]),
        },
    },
    # Primary Configs with Internal Batching
    {
        "name": "basic_multiple_tuple_target_with_internal_batching",
        "algorithms": [IntegratedGradients],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "internal_batch_size": 2,
        },
    },
    # NoiseTunnel
    {
        "name": "basic_multi_input_multi_target_nt",
        "algorithms": [
            IntegratedGradients,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            Saliency,
            GuidedBackprop,
            Deconvolution,
            LRP,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
            "nt_samples": 20,
            "stdevs": 0.0,
        },
        "noise_tunnel": True,
        "dp_delta": 0.01,
    },
    {
        "name": "basic_multiple_target_with_baseline_nt",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [0, 1, 1, 0],
            "nt_samples": 20,
            "stdevs": 0.0,
        },
        "noise_tunnel": True,
    },
    {
        "name": "basic_multiple_tuple_target_nt",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "nt_samples": 20,
            "stdevs": 0.0,
        },
        "noise_tunnel": True,
    },
    {
        "name": "basic_single_tensor_target_nt",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": torch.tensor([0]),
            "nt_samples": 20,
            "stdevs": 0.0,
        },
        "noise_tunnel": True,
    },
    {
        "name": "basic_multi_tensor_target_nt",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": torch.tensor([0, 1, 1, 0]),
            "nt_samples": 20,
            "stdevs": 0.0,
        },
        "noise_tunnel": True,
    },
    {
        "name": "basic_multi_tensor_target_batched_nt",
        "algorithms": [
            IntegratedGradients,
            Saliency,
            InputXGradient,
            FeatureAblation,
            DeepLift,
            GuidedBackprop,
            Deconvolution,
            LRP,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": torch.tensor([0, 1, 1, 0]),
            "nt_samples": 20,
            "nt_samples_batch_size": 2,
            "stdevs": 0.0,
        },
        "noise_tunnel": True,
    },
    # DeepLift SHAP
    {
        "name": "basic_dl_shap",
        "algorithms": [DeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(6, 3),
            "target": 0,
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_multi_input_dl_shap",
        "algorithms": [DeepLiftShap],
        "model": BasicModel_MultiLayer_MultiInput(),
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "baselines": (torch.randn(4, 3), torch.randn(4, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "target": 0,
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_multiple_target_dl_shap",
        "algorithms": [DeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(4, 3),
            "target": [0, 1, 1, 0],
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_multiple_tuple_target_dl_shap",
        "algorithms": [DeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_single_tensor_targe_dl_shap",
        "algorithms": [DeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(4, 3),
            "target": torch.tensor([0]),
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_multi_tensor_target_dl_shap",
        "algorithms": [DeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(4, 3),
            "target": torch.tensor([0, 1, 1, 0]),
        },
        "baseline_distr": True,
    },
    # Gradient SHAP
    {
        "name": "basic_multi_inp_with_single_baseline_grad_shap",
        "algorithms": [GradientShap],
        "model": BasicModel_MultiLayer_MultiInput(),
        "attribute_args": {
            "inputs": (torch.randn(6, 3), torch.randn(6, 3)),
            "baselines": (torch.randn(1, 3), torch.randn(1, 3)),
            "additional_forward_args": (torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
            "stdevs": 0.0,
            "n_samples": 2000,
        },
        "target_delta": 1.0,
        "dp_delta": 0.005,
        "baseline_distr": True,
    },
    {
        "name": "basic_multiple_target_with_single_baseline_grad_shap",
        "algorithms": [GradientShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(1, 3),
            "target": [0, 1, 1, 0],
            "n_samples": 800,
            "stdevs": 0.0,
        },
        "target_delta": 0.6,
        "baseline_distr": True,
    },
    {
        "name": "basic_multiple_tuple_target_with_single_baseline_grad_shap",
        "algorithms": [GradientShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.15 * torch.randn(1, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "n_samples": 2000,
            "stdevs": 0.0,
        },
        "target_delta": 0.6,
        "dp_delta": 0.003,
        "baseline_distr": True,
    },
    {
        "name": "basic_single_tensor_target_with_single_baseline_grad_shap",
        "algorithms": [GradientShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(1, 3),
            "target": torch.tensor([0]),
            "n_samples": 500,
            "stdevs": 0.0,
        },
        "target_delta": 0.6,
        "baseline_distr": True,
    },
    {
        "name": "basic_multi_tensor_target_with_single_baseline_grad_shap",
        "algorithms": [GradientShap],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(1, 3),
            "target": torch.tensor([0, 1, 1, 0]),
            "n_samples": 500,
            "stdevs": 0.0,
        },
        "target_delta": 0.6,
        "baseline_distr": True,
    },
    # Perturbation-Specific Configs
    {
        "name": "conv_with_perturbations_per_eval",
        "algorithms": [
            FeatureAblation,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
        ],
        "model": BasicModel_ConvNet(),
        "attribute_args": {
            "inputs": torch.arange(400).view(4, 1, 10, 10).float(),
            "target": 0,
            "perturbations_per_eval": 20,
        },
        "dp_delta": 0.008,
    },
    {
        "name": "basic_multiple_tuple_target_with_perturbations_per_eval",
        "algorithms": [
            FeatureAblation,
            ShapleyValueSampling,
            FeaturePermutation,
            Lime,
            KernelShap,
        ],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "perturbations_per_eval": 2,
        },
    },
    {
        "name": "conv_occlusion_with_perturbations_per_eval",
        "algorithms": [Occlusion],
        "model": BasicModel_ConvNet(),
        "attribute_args": {
            "inputs": torch.arange(400).view(4, 1, 10, 10).float(),
            "perturbations_per_eval": 8,
            "sliding_window_shapes": (1, 4, 2),
            "target": 0,
        },
    },
    {
        "name": "basic_multi_input_with_perturbations_per_eval_occlusion",
        "algorithms": [Occlusion],
        "model": ReLULinearModel(),
        "attribute_args": {
            "inputs": (torch.randn(4, 3), torch.randn(4, 3)),
            "perturbations_per_eval": 2,
            "sliding_window_shapes": ((2,), (1,)),
        },
    },
    {
        "name": "basic_multiple_tuple_target_occlusion",
        "algorithms": [Occlusion],
        "model": BasicModel_MultiLayer(),
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "sliding_window_shapes": (2,),
        },
    },
    # Layer Attribution Method Configs
    {
        "name": "conv_layer_single_target",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
            GuidedGradCam,
        ],
        "model": BasicModel_ConvNet(),
        "layer": "conv2",
        "attribute_args": {"inputs": 100 * torch.randn(4, 1, 10, 10), "target": 1},
    },
    {
        "name": "basic_layer_in_place",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
        ],
        "model": BasicModel_MultiLayer(inplace=True),
        "layer": "relu",
        "attribute_args": {"inputs": torch.randn(4, 3), "target": 0},
    },
    {
        "name": "basic_layer_multi_output",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
        ],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": "multi_relu",
        "attribute_args": {"inputs": torch.randn(4, 3), "target": 0},
    },
    {
        "name": "basic_layer_multi_input",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "target": 0,
        },
    },
    {
        "name": "basic_layer_multiple_target",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
        ],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {"inputs": torch.randn(4, 3), "target": [0, 1, 1, 0]},
    },
    {
        "name": "basic_layer_tensor_multiple_target",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
        ],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": torch.tensor([0, 1, 1, 0]),
        },
    },
    {
        "name": "basic_layer_multiple_tuple_target",
        "algorithms": [
            LayerConductance,
            LayerIntegratedGradients,
            LayerDeepLift,
            InternalInfluence,
            LayerFeatureAblation,
            LayerGradientXActivation,
            LayerGradCam,
        ],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
        },
    },
    {
        "name": "basic_layer_multiple_tuple_target_with_internal_batching",
        "algorithms": [LayerConductance, InternalInfluence, LayerIntegratedGradients],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "internal_batch_size": 2,
        },
    },
    {
        "name": "basic_layer_multi_input_with_internal_batching",
        "algorithms": [LayerConductance, InternalInfluence, LayerIntegratedGradients],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "target": 0,
            "internal_batch_size": 2,
        },
    },
    {
        "name": "basic_layer_multi_output_with_internal_batching",
        "algorithms": [LayerConductance, InternalInfluence, LayerIntegratedGradients],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": "multi_relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": 0,
            "internal_batch_size": 2,
        },
    },
    # Layer Perturbation
    {
        "name": "basic_layer_multi_input_with_perturbations_per_eval",
        "algorithms": [LayerFeatureAblation],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "target": 0,
            "perturbations_per_eval": 2,
        },
    },
    {
        "name": "basic_layer_multi_output_perturbations_per_eval",
        "algorithms": [LayerFeatureAblation],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": "multi_relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": 0,
            "perturbations_per_eval": 2,
        },
    },
    {
        "name": "conv_layer_with_perturbations_per_eval",
        "algorithms": [LayerFeatureAblation],
        "model": BasicModel_ConvNet(),
        "layer": "conv2",
        "attribute_args": {
            "inputs": 100 * torch.randn(4, 1, 10, 10),
            "target": 1,
            "perturbations_per_eval": 20,
        },
    },
    # Layer DeepLiftSHAP
    {
        "name": "relu_layer_multi_inp_dl_shap",
        "algorithms": [LayerDeepLiftShap],
        "model": ReLULinearModel(),
        "layer": "l3",
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "baselines": (2 * torch.randn(2, 3), 6 * torch.randn(2, 3)),
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_layer_multi_output_dl_shap",
        "algorithms": [LayerDeepLiftShap],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": "multi_relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": torch.randn(2, 3),
            "target": 0,
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_layer_multi_inp_multi_target_dl_shap",
        "algorithms": [LayerDeepLiftShap],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "baselines": (2 * torch.randn(11, 3), 6 * torch.randn(11, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_layer_multiple_target_dl_shap",
        "algorithms": [LayerDeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(6, 3),
            "target": [0, 1, 1, 0],
        },
        "baseline_distr": True,
    },
    # Layer Gradient SHAP
    {
        "name": "relu_layer_multi_inp_grad_shap",
        "algorithms": [LayerGradientShap],
        "model": ReLULinearModel(),
        "layer": "l3",
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "baselines": (2 * torch.randn(2, 3), 6 * torch.randn(2, 3)),
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_layer_multi_output_grad_shap",
        "algorithms": [LayerGradientShap],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": "multi_relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": torch.randn(2, 3),
            "target": 0,
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_layer_multi_inp_multi_target_grad_shap",
        "algorithms": [LayerGradientShap],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (torch.randn(6, 3), torch.randn(6, 3)),
            "baselines": (torch.randn(2, 3), torch.randn(2, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
            "n_samples": 1000,
        },
        "baseline_distr": True,
        "target_delta": 0.6,
    },
    # Neuron Attribution Method Configs
    {
        "name": "basic_neuron",
        "algorithms": [
            NeuronGradient,
            NeuronIntegratedGradients,
            NeuronGuidedBackprop,
            NeuronDeconvolution,
            NeuronDeepLift,
            NeuronFeatureAblation,
        ],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {"inputs": torch.randn(4, 3), "neuron_selector": 3},
    },
    {
        "name": "conv_neuron",
        "algorithms": [
            NeuronGradient,
            NeuronIntegratedGradients,
            NeuronGuidedBackprop,
            NeuronDeconvolution,
            NeuronDeepLift,
            NeuronFeatureAblation,
        ],
        "model": BasicModel_ConvNet(),
        "layer": "conv2",
        "attribute_args": {
            "inputs": 100 * torch.randn(4, 1, 10, 10),
            "neuron_selector": (0, 1, 0),
        },
    },
    {
        "name": "basic_neuron_multi_input",
        "algorithms": [
            NeuronGradient,
            NeuronIntegratedGradients,
            NeuronGuidedBackprop,
            NeuronDeconvolution,
            NeuronDeepLift,
            NeuronFeatureAblation,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "neuron_selector": (3,),
        },
    },
    # Neuron Conductance (with target)
    {
        "name": "basic_neuron_single_target",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": 1,
            "neuron_selector": 3,
        },
    },
    {
        "name": "basic_neuron_multiple_target",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [0, 1, 1, 0],
            "neuron_selector": 3,
        },
    },
    {
        "name": "conv_neuron_single_target",
        "algorithms": [NeuronConductance],
        "model": BasicModel_ConvNet(),
        "layer": "conv2",
        "attribute_args": {
            "inputs": 100 * torch.randn(4, 1, 10, 10),
            "target": 1,
            "neuron_selector": (0, 1, 0),
        },
    },
    {
        "name": "basic_neuron_multi_input_multi_target",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
            "neuron_selector": 3,
        },
    },
    {
        "name": "basic_neuron_tensor_multiple_target",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": torch.tensor([0, 1, 1, 0]),
            "neuron_selector": 3,
        },
    },
    {
        "name": "basic_neuron_multiple_tuple_target",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "neuron_selector": 3,
        },
    },
    # Neuron Conductance with Internal Batching
    {
        "name": "basic_neuron_multiple_tuple_target_with_internal_batching",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "target": [(1, 0, 0), (0, 1, 1), (1, 1, 1), (0, 0, 0)],
            "additional_forward_args": (None, True),
            "internal_batch_size": 2,
            "neuron_selector": 3,
        },
    },
    {
        "name": "basic_neuron_multi_input_multi_target_with_internal_batching",
        "algorithms": [NeuronConductance],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "target": [0, 1, 1, 0, 0, 1],
            "internal_batch_size": 2,
            "neuron_selector": 3,
        },
    },
    # Neuron Gradient SHAP
    {
        "name": "basic_neuron_grad_shap",
        "algorithms": [NeuronGradientShap],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": torch.randn(1, 3),
            "neuron_selector": 3,
        },
        "target_delta": 0.6,
        "baseline_distr": True,
    },
    {
        "name": "basic_neuron_multi_inp_grad_shap",
        "algorithms": [NeuronGradientShap],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(6, 3), 5 * torch.randn(6, 3)),
            "baselines": (10 * torch.randn(1, 3), 5 * torch.randn(1, 3)),
            "additional_forward_args": (2 * torch.randn(6, 3), 5),
            "neuron_selector": 3,
        },
        "target_delta": 0.6,
        "baseline_distr": True,
    },
    # Neuron DeepLift SHAP
    {
        "name": "basic_neuron_dl_shap",
        "algorithms": [NeuronDeepLiftShap],
        "model": BasicModel_MultiLayer(),
        "layer": "relu",
        "attribute_args": {
            "inputs": torch.randn(4, 3),
            "baselines": 0.5 * torch.randn(6, 3),
            "neuron_selector": (3,),
        },
        "baseline_distr": True,
    },
    {
        "name": "basic_neuron_multi_input_dl_shap",
        "algorithms": [NeuronDeepLiftShap],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "baselines": (torch.randn(4, 3), torch.randn(4, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "neuron_selector": 3,
        },
        "baseline_distr": True,
    },
    # Neuron Feature Ablation
    {
        "name": "conv_neuron_with_perturbations_per_eval",
        "algorithms": [NeuronFeatureAblation],
        "model": BasicModel_ConvNet(),
        "layer": "conv2",
        "attribute_args": {
            "inputs": torch.arange(400).view(4, 1, 10, 10).float(),
            "perturbations_per_eval": 20,
            "neuron_selector": (0, 1, 0),
        },
    },
    {
        "name": "basic_neuron_multiple_input_with_baselines_and_perturbations_per_eval",
        "algorithms": [NeuronFeatureAblation],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "baselines": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "neuron_selector": (3,),
            "perturbations_per_eval": 2,
        },
    },
    # Neuron Attribution with Functional Selector
    {
        "name": "basic_neuron_multi_input_function_selector",
        "algorithms": [
            NeuronGradient,
            NeuronIntegratedGradients,
            NeuronGuidedBackprop,
            NeuronDeconvolution,
            NeuronDeepLift,
            NeuronFeatureAblation,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.relu",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "neuron_selector": lambda x: torch.sum(x, 1),
        },
    },
    # Neuron Attribution with slice Selector
    {
        "name": "conv_neuron_slice_selector",
        "algorithms": [
            NeuronGradient,
            NeuronIntegratedGradients,
            NeuronGuidedBackprop,
            NeuronDeconvolution,
            NeuronDeepLift,
            NeuronFeatureAblation,
        ],
        "model": BasicModel_ConvNet(),
        "layer": "conv2",
        "attribute_args": {
            "inputs": 100 * torch.randn(4, 1, 10, 10),
            "neuron_selector": (slice(0, 2, 1), 1, slice(0, 2, 1)),
        },
    },
    # Layer Attribution with Multiple Layers
    {
        "name": "basic_activation_multi_layer_multi_output",
        "algorithms": [LayerActivation],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": ["multi_relu", "linear1", "linear0"],
        "attribute_args": {"inputs": torch.randn(4, 3)},
    },
    {
        "name": "basic_gradient_multi_layer_multi_output",
        "algorithms": [LayerGradientXActivation],
        "model": BasicModel_MultiLayer(multi_input_module=True),
        "layer": ["multi_relu", "linear1", "linear0"],
        "attribute_args": {"inputs": torch.randn(4, 3), "target": 0},
    },
    {
        "name": "basic_layer_ig_multi_layer_multi_output",
        "algorithms": [LayerIntegratedGradients],
        "model": BasicModel_MultiLayer_TrueMultiInput(),
        "layer": ["m1", "m234"],
        "attribute_args": {
            "inputs": (
                torch.randn(5, 3),
                torch.randn(5, 3),
                torch.randn(5, 3),
                torch.randn(5, 3),
            ),
            "target": 0,
        },
    },
    {
        "name": "basic_layer_ig_multi_layer_multi_output_with_input_wrapper",
        "algorithms": [LayerIntegratedGradients],
        "model": ModelInputWrapper(BasicModel_MultiLayer_TrueMultiInput()),
        "layer": ["module.m1", "module.m234"],
        "attribute_args": {
            "inputs": (
                torch.randn(5, 3),
                torch.randn(5, 3),
                torch.randn(5, 3),
                torch.randn(5, 3),
            ),
            "target": 0,
        },
    },
    # Layer LRP
    {
        "name": "basic_layer_lrp",
        "algorithms": [
            LayerLRP,
        ],
        "model": BasicModel_MultiLayer(),
        "layer": "linear2",
        "attribute_args": {"inputs": torch.randn(4, 3), "target": 0},
    },
    {
        "name": "basic_layer_lrp_multi_input",
        "algorithms": [
            LayerLRP,
        ],
        "model": BasicModel_MultiLayer_MultiInput(),
        "layer": "model.linear1",
        "attribute_args": {
            "inputs": (10 * torch.randn(12, 3), 5 * torch.randn(12, 3)),
            "additional_forward_args": (2 * torch.randn(12, 3), 5),
            "target": 0,
        },
        "dp_delta": 0.0002,
    },
]



#!/usr/bin/env python3
import typing
from typing import Any, Callable, Tuple, Union

import numpy as np
import torch
from captum._utils.common import _is_tuple
from captum._utils.typing import (
    BaselineType,
    Literal,
    TargetType,
    Tensor,
    TensorOrTupleOfTensorsGeneric,
)
from captum.attr._core.noise_tunnel import NoiseTunnel
from captum.attr._utils.attribution import GradientAttribution
from captum.attr._utils.common import (
    _compute_conv_delta_and_format_attrs,
    _format_callable_baseline,
    _format_input_baseline,
)
from captum.log import log_usage


class GradientShap(GradientAttribution):
    r"""
    Implements gradient SHAP based on the implementation from SHAP's primary
    author. For reference, please view the original
    `implementation
    <https://github.com/slundberg/shap#deep-learning-example-with-gradientexplainer-tensorflowkeraspytorch-models>`_
    and the paper: `A Unified Approach to Interpreting Model Predictions
    <https://papers.nips.cc/paper/7062-a-unified-approach-to-interpreting-model-predictions>`_

    GradientShap approximates SHAP values by computing the expectations of
    gradients by randomly sampling from the distribution of baselines/references.
    It adds white noise to each input sample `n_samples` times, selects a
    random baseline from baselines' distribution and a random point along the
    path between the baseline and the input, and computes the gradient of outputs
    with respect to those selected random points. The final SHAP values represent
    the expected values of gradients * (inputs - baselines).

    GradientShap makes an assumption that the input features are independent
    and that the explanation model is linear, meaning that the explanations
    are modeled through the additive composition of feature effects.
    Under those assumptions, SHAP value can be approximated as the expectation
    of gradients that are computed for randomly generated `n_samples` input
    samples after adding gaussian noise `n_samples` times to each input for
    different baselines/references.

    In some sense it can be viewed as an approximation of integrated gradients
    by computing the expectations of gradients for different baselines.

    Current implementation uses Smoothgrad from :class:`.NoiseTunnel` in order to
    randomly draw samples from the distribution of baselines, add noise to input
    samples and compute the expectation (smoothgrad).
    """

    def __init__(self, forward_func: Callable, multiply_by_inputs: bool = True) -> None:
        r"""
        Args:

            forward_func (Callable): The forward function of the model or
                       any modification of it.
            multiply_by_inputs (bool, optional): Indicates whether to factor
                    model inputs' multiplier in the final attribution scores.
                    In the literature this is also known as local vs global
                    attribution. If inputs' multiplier isn't factored in
                    then this type of attribution method is also called local
                    attribution. If it is, then that type of attribution
                    method is called global.
                    More detailed can be found here:
                    https://arxiv.org/abs/1711.06104

                    In case of gradient shap, if `multiply_by_inputs`
                    is set to True, the sensitivity scores of scaled inputs
                    are being multiplied by (inputs - baselines).
        """
        GradientAttribution.__init__(self, forward_func)
        self._multiply_by_inputs = multiply_by_inputs

    @typing.overload
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: Union[
            TensorOrTupleOfTensorsGeneric, Callable[..., TensorOrTupleOfTensorsGeneric]
        ],
        n_samples: int = 5,
        stdevs: Union[float, Tuple[float, ...]] = 0.0,
        target: TargetType = None,
        additional_forward_args: Any = None,
        *,
        return_convergence_delta: Literal[True],
    ) -> Tuple[TensorOrTupleOfTensorsGeneric, Tensor]:
        ...

    @typing.overload
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: Union[
            TensorOrTupleOfTensorsGeneric, Callable[..., TensorOrTupleOfTensorsGeneric]
        ],
        n_samples: int = 5,
        stdevs: Union[float, Tuple[float, ...]] = 0.0,
        target: TargetType = None,
        additional_forward_args: Any = None,
        return_convergence_delta: Literal[False] = False,
    ) -> TensorOrTupleOfTensorsGeneric:
        ...

    @log_usage()
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: Union[
            TensorOrTupleOfTensorsGeneric, Callable[..., TensorOrTupleOfTensorsGeneric]
        ],
        n_samples: int = 5,
        stdevs: Union[float, Tuple[float, ...]] = 0.0,
        target: TargetType = None,
        additional_forward_args: Any = None,
        return_convergence_delta: bool = False,
    ) -> Union[
        TensorOrTupleOfTensorsGeneric, Tuple[TensorOrTupleOfTensorsGeneric, Tensor]
    ]:
        r"""
        Args:

            inputs (Tensor or tuple[Tensor, ...]): Input for which SHAP attribution
                        values are computed. If `forward_func` takes a single
                        tensor as input, a single input tensor should be provided.
                        If `forward_func` takes multiple tensors as input, a tuple
                        of the input tensors should be provided. It is assumed
                        that for all given input tensors, dimension 0 corresponds
                        to the number of examples, and if multiple input tensors
                        are provided, the examples must be aligned appropriately.
            baselines (Tensor, tuple[Tensor, ...], or Callable):
                        Baselines define the starting point from which expectation
                        is computed and can be provided as:

                        - a single tensor, if inputs is a single tensor, with
                          the first dimension equal to the number of examples
                          in the baselines' distribution. The remaining dimensions
                          must match with input tensor's dimension starting from
                          the second dimension.

                        - a tuple of tensors, if inputs is a tuple of tensors,
                          with the first dimension of any tensor inside the tuple
                          equal to the number of examples in the baseline's
                          distribution. The remaining dimensions must match
                          the dimensions of the corresponding input tensor
                          starting from the second dimension.

                        - callable function, optionally takes `inputs` as an
                          argument and either returns a single tensor
                          or a tuple of those.

                        It is recommended that the number of samples in the baselines'
                        tensors is larger than one.
            n_samples (int, optional): The number of randomly generated examples
                        per sample in the input batch. Random examples are
                        generated by adding gaussian random noise to each sample.
                        Default: `5` if `n_samples` is not provided.
            stdevs    (float or tuple of float, optional): The standard deviation
                        of gaussian noise with zero mean that is added to each
                        input in the batch. If `stdevs` is a single float value
                        then that same value is used for all inputs. If it is
                        a tuple, then it must have the same length as the inputs
                        tuple. In this case, each stdev value in the stdevs tuple
                        corresponds to the input with the same index in the inputs
                        tuple.
                        Default: 0.0
            target (int, tuple, Tensor, or list, optional): Output indices for
                        which gradients are computed (for classification cases,
                        this is usually the target class).
                        If the network returns a scalar value per example,
                        no target index is necessary.
                        For general 2D outputs, targets can be either:

                        - a single integer or a tensor containing a single
                          integer, which is applied to all input examples

                        - a list of integers or a 1D tensor, with length matching
                          the number of examples in inputs (dim 0). Each integer
                          is applied as the target for the corresponding example.

                        For outputs with > 2 dimensions, targets can be either:

                        - A single tuple, which contains #output_dims - 1
                          elements. This target index is applied to all examples.

                        - A list of tuples with length equal to the number of
                          examples in inputs (dim 0), and each tuple containing
                          #output_dims - 1 elements. Each tuple is applied as the
                          target for the corresponding example.

                        Default: None
            additional_forward_args (Any, optional): If the forward function
                        requires additional arguments other than the inputs for
                        which attributions should not be computed, this argument
                        can be provided. It can contain a tuple of ND tensors or
                        any arbitrary python type of any shape.
                        In case of the ND tensor the first dimension of the
                        tensor must correspond to the batch size. It will be
                        repeated for each `n_steps` for each randomly generated
                        input sample.
                        Note that the gradients are not computed with respect
                        to these arguments.
                        Default: None
            return_convergence_delta (bool, optional): Indicates whether to return
                        convergence delta or not. If `return_convergence_delta`
                        is set to True convergence delta will be returned in
                        a tuple following attributions.
                        Default: False
        Returns:
            **attributions** or 2-element tuple of **attributions**, **delta**:
            - **attributions** (*Tensor* or *tuple[Tensor, ...]*):
                        Attribution score computed based on GradientSHAP with respect
                        to each input feature. Attributions will always be
                        the same size as the provided inputs, with each value
                        providing the attribution of the corresponding input index.
                        If a single tensor is provided as inputs, a single tensor is
                        returned. If a tuple is provided for inputs, a tuple of
                        corresponding sized tensors is returned.
            - **delta** (*Tensor*, returned if return_convergence_delta=True):
                        This is computed using the property that the total
                        sum of forward_func(inputs) - forward_func(baselines)
                        must be very close to the total sum of the attributions
                        based on GradientSHAP.
                        Delta is calculated for each example in the input after adding
                        `n_samples` times gaussian noise to each of them. Therefore,
                        the dimensionality of the deltas tensor is equal to the
                        `number of examples in the input` * `n_samples`
                        The deltas are ordered by each input example and `n_samples`
                        noisy samples generated for it.

        Examples::

            >>> # ImageClassifier takes a single input tensor of images Nx3x32x32,
            >>> # and returns an Nx10 tensor of class probabilities.
            >>> net = ImageClassifier()
            >>> gradient_shap = GradientShap(net)
            >>> input = torch.randn(3, 3, 32, 32, requires_grad=True)
            >>> # choosing baselines randomly
            >>> baselines = torch.randn(20, 3, 32, 32)
            >>> # Computes gradient shap for the input
            >>> # Attribution size matches input size: 3x3x32x32
            >>> attribution = gradient_shap.attribute(input, baselines,
                                                                target=5)

        """
        # since `baselines` is a distribution, we can generate it using a function
        # rather than passing it as an input argument
        baselines = _format_callable_baseline(baselines, inputs)
        assert isinstance(baselines[0], torch.Tensor), (
            "Baselines distribution has to be provided in a form "
            "of a torch.Tensor {}.".format(baselines[0])
        )

        input_min_baseline_x_grad = InputBaselineXGradient(
            self.forward_func, self.multiplies_by_inputs
        )
        input_min_baseline_x_grad.gradient_func = self.gradient_func

        nt = NoiseTunnel(input_min_baseline_x_grad)

        # NOTE: using attribute.__wrapped__ to not log
        attributions = nt.attribute.__wrapped__(
            nt,  # self
            inputs,
            nt_type="smoothgrad",
            nt_samples=n_samples,
            stdevs=stdevs,
            draw_baseline_from_distrib=True,
            baselines=baselines,
            target=target,
            additional_forward_args=additional_forward_args,
            return_convergence_delta=return_convergence_delta,
        )

        return attributions

    def has_convergence_delta(self) -> bool:
        return True

    @property
    def multiplies_by_inputs(self):
        return self._multiply_by_inputs


class InputBaselineXGradient(GradientAttribution):
    def __init__(self, forward_func: Callable, multiply_by_inputs=True) -> None:
        r"""
        Args:

            forward_func (Callable): The forward function of the model or
                        any modification of it.
            multiply_by_inputs (bool, optional): Indicates whether to factor
                        model inputs' multiplier in the final attribution scores.
                        In the literature this is also known as local vs global
                        attribution. If inputs' multiplier isn't factored in
                        then this type of attribution method is also called local
                        attribution. If it is, then that type of attribution
                        method is called global.
                        More detailed can be found here:
                        https://arxiv.org/abs/1711.06104

                        In case of gradient shap, if `multiply_by_inputs`
                        is set to True, the sensitivity scores of scaled inputs
                        are being multiplied by (inputs - baselines).

        """
        GradientAttribution.__init__(self, forward_func)
        self._multiply_by_inputs = multiply_by_inputs

    @typing.overload
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: BaselineType = None,
        target: TargetType = None,
        additional_forward_args: Any = None,
        *,
        return_convergence_delta: Literal[True],
    ) -> Tuple[TensorOrTupleOfTensorsGeneric, Tensor]:
        ...

    @typing.overload
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: BaselineType = None,
        target: TargetType = None,
        additional_forward_args: Any = None,
        return_convergence_delta: Literal[False] = False,
    ) -> TensorOrTupleOfTensorsGeneric:
        ...

    @log_usage()
    def attribute(  # type: ignore
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: BaselineType = None,
        target: TargetType = None,
        additional_forward_args: Any = None,
        return_convergence_delta: bool = False,
    ) -> Union[
        TensorOrTupleOfTensorsGeneric, Tuple[TensorOrTupleOfTensorsGeneric, Tensor]
    ]:
        # Keeps track whether original input is a tuple or not before
        # converting it into a tuple.
        is_inputs_tuple = _is_tuple(inputs)
        inputs, baselines = _format_input_baseline(inputs, baselines)

        rand_coefficient = torch.tensor(
            np.random.uniform(0.0, 1.0, inputs[0].shape[0]),
            device=inputs[0].device,
            dtype=inputs[0].dtype,
        )

        input_baseline_scaled = tuple(
            _scale_input(input, baseline, rand_coefficient)
            for input, baseline in zip(inputs, baselines)
        )
        grads = self.gradient_func(
            self.forward_func, input_baseline_scaled, target, additional_forward_args
        )

        if self.multiplies_by_inputs:
            input_baseline_diffs = tuple(
                input - baseline for input, baseline in zip(inputs, baselines)
            )
            attributions = tuple(
                input_baseline_diff * grad
                for input_baseline_diff, grad in zip(input_baseline_diffs, grads)
            )
        else:
            attributions = grads

        return _compute_conv_delta_and_format_attrs(
            self,
            return_convergence_delta,
            attributions,
            baselines,
            inputs,
            additional_forward_args,
            target,
            is_inputs_tuple,
        )

    def has_convergence_delta(self) -> bool:
        return True

    @property
    def multiplies_by_inputs(self):
        return self._multiply_by_inputs


def _scale_input(
    input: Tensor, baseline: Union[Tensor, int, float], rand_coefficient: Tensor
) -> Tensor:
    # batch size
    bsz = input.shape[0]
    inp_shape_wo_bsz = input.shape[1:]
    inp_shape = (bsz,) + tuple([1] * len(inp_shape_wo_bsz))

    # expand and reshape the indices
    rand_coefficient = rand_coefficient.view(inp_shape)

    input_baseline_scaled = (
        rand_coefficient * input + (1.0 - rand_coefficient) * baseline
    ).requires_grad_()
    return input_baseline_scaled



from captum.insights.attr_vis.widget._version import __version__, version_info  # noqa
from captum.insights.attr_vis.widget.widget import *  # noqa


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "static",
            "dest": "jupyter-captum-insights",
            "require": "jupyter-captum-insights/extension",
        }
    ]



#!/usr/bin/env python3
import torch
from captum.attr import CommonStats, Summarizer
from tests.helpers.basic import BaseTest


class Test(BaseTest):
    def test_single_input(self) -> None:
        size = (2, 3)
        summarizer = Summarizer(stats=CommonStats())
        for _ in range(10):
            attrs = torch.randn(size)
            summarizer.update(attrs)

        summ = summarizer.summary
        self.assertIsNotNone(summ)
        self.assertTrue(isinstance(summ, dict))

        for k in summ:
            self.assertTrue(summ[k].size() == size)

    def test_multi_input(self) -> None:
        size1 = (10, 5, 5)
        size2 = (3, 5)

        summarizer = Summarizer(stats=CommonStats())
        for _ in range(10):
            a1 = torch.randn(size1)
            a2 = torch.randn(size2)
            summarizer.update((a1, a2))

        summ = summarizer.summary
        self.assertIsNotNone(summ)
        self.assertTrue(len(summ) == 2)
        self.assertTrue(isinstance(summ[0], dict))
        self.assertTrue(isinstance(summ[1], dict))

        for k in summ[0]:
            self.assertTrue(summ[0][k].size() == size1)
            self.assertTrue(summ[1][k].size() == size2)



from typing import Callable, cast, List, Optional

import torch.nn as nn
from captum._utils.models.model import Model
from torch import Tensor
from torch.utils.data import DataLoader


class LinearModel(nn.Module, Model):
    SUPPORTED_NORMS: List[Optional[str]] = [None, "batch_norm", "layer_norm"]

    def __init__(self, train_fn: Callable, **kwargs) -> None:
        r"""
        Constructs a linear model with a training function and additional
        construction arguments that will be sent to
        `self._construct_model_params` after a `self.fit` is called. Please note
        that this assumes the `self.train_fn` will call
        `self._construct_model_params`.

        Please note that this is an experimental feature.

        Args:
            train_fn (Callable)
                The function to train with. See
                `captum._utils.models.linear_model.train.sgd_train_linear_model`
                and
                `captum._utils.models.linear_model.train.sklearn_train_linear_model`
                for examples
            kwargs
                Any additional keyword arguments to send to
                `self._construct_model_params` once a `self.fit` is called.
        """
        super().__init__()

        self.norm: Optional[nn.Module] = None
        self.linear: Optional[nn.Linear] = None
        self.train_fn = train_fn
        self.construct_kwargs = kwargs

    def _construct_model_params(
        self,
        in_features: Optional[int] = None,
        out_features: Optional[int] = None,
        norm_type: Optional[str] = None,
        affine_norm: bool = False,
        bias: bool = True,
        weight_values: Optional[Tensor] = None,
        bias_value: Optional[Tensor] = None,
        classes: Optional[Tensor] = None,
    ):
        r"""
        Lazily initializes a linear model. This will be called for you in a
        train method.

        Args:
            in_features (int):
                The number of input features
            output_features (int):
                The number of output features.
            norm_type (str, optional):
                The type of normalization that can occur. Please assign this
                to one of `PyTorchLinearModel.SUPPORTED_NORMS`.
            affine_norm (bool):
                Whether or not to learn an affine transformation of the
                normalization parameters used.
            bias (bool):
                Whether to add a bias term. Not needed if normalized input.
            weight_values (Tensor, optional):
                The values to initialize the linear model with. This must be a
                1D or 2D tensor, and of the form `(num_outputs, num_features)` or
                `(num_features,)`. Additionally, if this is provided you need not
                to provide `in_features` or `out_features`.
            bias_value (Tensor, optional):
                The bias value to initialize the model with.
            classes (Tensor, optional):
                The list of prediction classes supported by the model in case it
                performs classificaton. In case of regression it is set to None.
                Default: None
        """
        if norm_type not in LinearModel.SUPPORTED_NORMS:
            raise ValueError(
                f"{norm_type} not supported. Please use {LinearModel.SUPPORTED_NORMS}"
            )

        if weight_values is not None:
            in_features = weight_values.shape[-1]
            out_features = (
                1 if len(weight_values.shape) == 1 else weight_values.shape[0]
            )

        if in_features is None or out_features is None:
            raise ValueError(
                "Please provide `in_features` and `out_features` or `weight_values`"
            )

        if norm_type == "batch_norm":
            self.norm = nn.BatchNorm1d(in_features, eps=1e-8, affine=affine_norm)
        elif norm_type == "layer_norm":
            self.norm = nn.LayerNorm(
                in_features, eps=1e-8, elementwise_affine=affine_norm
            )
        else:
            self.norm = None

        self.linear = nn.Linear(in_features, out_features, bias=bias)

        if weight_values is not None:
            self.linear.weight.data = weight_values

        if bias_value is not None:
            if not bias:
                raise ValueError("`bias_value` is not None and bias is False")

            self.linear.bias.data = bias_value

        if classes is not None:
            self.linear.classes = classes

    def fit(self, train_data: DataLoader, **kwargs):
        r"""
        Calls `self.train_fn`
        """
        return self.train_fn(
            self,
            dataloader=train_data,
            construct_kwargs=self.construct_kwargs,
            **kwargs,
        )

    def forward(self, x: Tensor) -> Tensor:
        assert self.linear is not None
        if self.norm is not None:
            x = self.norm(x)
        return self.linear(x)

    def representation(self) -> Tensor:
        r"""
        Returns a tensor which describes the hyper-plane input space. This does
        not include the bias. For bias/intercept, please use `self.bias`
        """
        assert self.linear is not None
        return self.linear.weight.detach()

    def bias(self) -> Optional[Tensor]:
        r"""
        Returns the bias of the linear model
        """
        if self.linear is None or self.linear.bias is None:
            return None
        return self.linear.bias.detach()

    def classes(self) -> Optional[Tensor]:
        if self.linear is None or self.linear.classes is None:
            return None
        return cast(Tensor, self.linear.classes).detach()


class SGDLinearModel(LinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class. Construct a a `LinearModel` with the
        `sgd_train_linear_model` as the train method

        Args:
            kwargs
                Arguments send to `self._construct_model_params` after
                `self.fit` is called. Please refer to that method for parameter
                documentation.
        """
        # avoid cycles
        from captum._utils.models.linear_model.train import sgd_train_linear_model

        super().__init__(train_fn=sgd_train_linear_model, **kwargs)


class SGDLasso(SGDLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class to train a `LinearModel` with SGD
        (`sgd_train_linear_model`) whilst setting appropriate parameters to
        optimize for ridge regression loss. This optimizes L2 loss + alpha * L1
        regularization.

        Please note that with SGD it is not guaranteed that weights will
        converge to 0.
        """
        super().__init__(**kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        # avoid cycles
        from captum._utils.models.linear_model.train import l2_loss

        return super().fit(train_data=train_data, loss_fn=l2_loss, reg_term=1, **kwargs)


class SGDRidge(SGDLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class to train a `LinearModel` with SGD
        (`sgd_train_linear_model`) whilst setting appropriate parameters to
        optimize for ridge regression loss. This optimizes L2 loss + alpha *
        L2 regularization.
        """
        super().__init__(**kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        # avoid cycles
        from captum._utils.models.linear_model.train import l2_loss

        return super().fit(train_data=train_data, loss_fn=l2_loss, reg_term=2, **kwargs)


class SGDLinearRegression(SGDLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class to train a `LinearModel` with SGD
        (`sgd_train_linear_model`). For linear regression this assigns the loss
        to L2 and no regularization.
        """
        super().__init__(**kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        # avoid cycles
        from captum._utils.models.linear_model.train import l2_loss

        return super().fit(
            train_data=train_data, loss_fn=l2_loss, reg_term=None, **kwargs
        )


class SkLearnLinearModel(LinearModel):
    def __init__(self, sklearn_module: str, **kwargs) -> None:
        r"""
        Factory class to construct a `LinearModel` with sklearn training method.

        Please note that this assumes:

        0. You have sklearn and numpy installed
        1. The dataset can fit into memory

        SkLearn support does introduce some slight overhead as we convert the
        tensors to numpy and then convert the resulting trained model to a
        `LinearModel` object. However, this conversion should be negligible.

        Args:
            sklearn_module
                The module under sklearn to construct and use for training, e.g.
                use "svm.LinearSVC" for an SVM or "linear_model.Lasso" for Lasso.

                There are factory classes defined for you for common use cases,
                such as `SkLearnLasso`.
            kwargs
                The kwargs to pass to the construction of the sklearn model
        """
        # avoid cycles
        from captum._utils.models.linear_model.train import sklearn_train_linear_model

        super().__init__(train_fn=sklearn_train_linear_model, **kwargs)

        self.sklearn_module = sklearn_module

    def fit(self, train_data: DataLoader, **kwargs):
        r"""
        Args:
            train_data
                Train data to use
            kwargs
                Arguments to feed to `.fit` method for sklearn
        """
        return super().fit(
            train_data=train_data, sklearn_trainer=self.sklearn_module, **kwargs
        )


class SkLearnLasso(SkLearnLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class. Trains a `LinearModel` model with
        `sklearn.linear_model.Lasso`. You will need sklearn version >= 0.23 to
        support sample weights.
        """
        super().__init__(sklearn_module="linear_model.Lasso", **kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        return super().fit(train_data=train_data, **kwargs)


class SkLearnRidge(SkLearnLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class. Trains a model with `sklearn.linear_model.Ridge`.

        Any arguments provided to the sklearn constructor can be provided
        as kwargs here.
        """
        super().__init__(sklearn_module="linear_model.Ridge", **kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        return super().fit(train_data=train_data, **kwargs)


class SkLearnLinearRegression(SkLearnLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class. Trains a model with `sklearn.linear_model.LinearRegression`.

        Any arguments provided to the sklearn constructor can be provided
        as kwargs here.
        """
        super().__init__(sklearn_module="linear_model.LinearRegression", **kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        return super().fit(train_data=train_data, **kwargs)


class SkLearnLogisticRegression(SkLearnLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class. Trains a model with `sklearn.linear_model.LogisticRegression`.

        Any arguments provided to the sklearn constructor can be provided
        as kwargs here.
        """
        super().__init__(sklearn_module="linear_model.LogisticRegression", **kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        return super().fit(train_data=train_data, **kwargs)


class SkLearnSGDClassifier(SkLearnLinearModel):
    def __init__(self, **kwargs) -> None:
        r"""
        Factory class. Trains a model with `sklearn.linear_model.SGDClassifier(`.

        Any arguments provided to the sklearn constructor can be provided
        as kwargs here.
        """
        super().__init__(sklearn_module="linear_model.SGDClassifier", **kwargs)

    def fit(self, train_data: DataLoader, **kwargs):
        return super().fit(train_data=train_data, **kwargs)



import tempfile
from typing import Callable

import torch.nn as nn
from captum.influence._core.tracincp import TracInCP
from captum.influence._core.tracincp_fast_rand_proj import TracInCPFast

from parameterized import parameterized
from tests.helpers.basic import BaseTest
from tests.influence._utils.common import (
    build_test_name_func,
    DataInfluenceConstructor,
    get_random_model_and_data,
)


class TestTracinValidator(BaseTest):

    param_list = []
    for reduction, constr in [
        (
            "none",
            DataInfluenceConstructor(TracInCP, name="TracInCP"),
        ),
        (
            "mean",
            DataInfluenceConstructor(
                TracInCPFast,
                name="TracInCpFast",
            ),
        ),
    ]:
        param_list.append((reduction, constr))

    @parameterized.expand(
        param_list,
        name_func=build_test_name_func(),
    )
    def test_tracin_require_inputs_dataset(
        self,
        reduction,
        tracin_constructor: Callable,
    ) -> None:
        """
        This test verifies that tracinCP and tracinCPFast
        influence methods required `inputs_dataset`.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            (
                net,
                train_dataset,
                test_samples,
                test_labels,
            ) = get_random_model_and_data(tmpdir, unpack_inputs=False)

            criterion = nn.MSELoss(reduction=reduction)

            tracin = tracin_constructor(
                net,
                train_dataset,
                tmpdir,
                loss_fn=criterion,
                batch_size=1,
            )
            with self.assertRaisesRegex(AssertionError, "required."):
                tracin.influence(None, k=None)



#!/usr/bin/env python3
from captum.concept._core.cav import CAV  # noqa
from captum.concept._core.concept import Concept, ConceptInterpreter  # noqa
from captum.concept._core.tcav import TCAV  # noqa
from captum.concept._utils.classifier import Classifier, DefaultClassifier  # noqa



import time
import warnings
from typing import Any, Callable, Dict, List, Optional

import torch
import torch.nn as nn
from captum._utils.models.linear_model.model import LinearModel
from torch.utils.data import DataLoader


def l2_loss(x1, x2, weights=None):
    if weights is None:
        return torch.mean((x1 - x2) ** 2) / 2.0
    else:
        return torch.sum((weights / weights.norm(p=1)) * ((x1 - x2) ** 2)) / 2.0


def sgd_train_linear_model(
    model: LinearModel,
    dataloader: DataLoader,
    construct_kwargs: Dict[str, Any],
    max_epoch: int = 100,
    reduce_lr: bool = True,
    initial_lr: float = 0.01,
    alpha: float = 1.0,
    loss_fn: Callable = l2_loss,
    reg_term: Optional[int] = 1,
    patience: int = 10,
    threshold: float = 1e-4,
    running_loss_window: Optional[int] = None,
    device: Optional[str] = None,
    init_scheme: str = "zeros",
    debug: bool = False,
) -> Dict[str, float]:
    r"""
    Trains a linear model with SGD. This will continue to iterate your
    dataloader until we converged to a solution or alternatively until we have
    exhausted `max_epoch`.

    Convergence is defined by the loss not changing by `threshold` amount for
    `patience` number of iterations.

    Args:
        model
            The model to train
        dataloader
            The data to train it with. We will assume the dataloader produces
            either pairs or triples of the form (x, y) or (x, y, w). Where x and
            y are typical pairs for supervised learning and w is a weight
            vector.

            We will call `model._construct_model_params` with construct_kwargs
            and the input features set to `x.shape[1]` (`x.shape[0]` corresponds
            to the batch size). We assume that `len(x.shape) == 2`, i.e. the
            tensor is flat. The number of output features will be set to
            y.shape[1] or 1 (if `len(y.shape) == 1`); we require `len(y.shape)
            <= 2`.
        max_epoch
            The maximum number of epochs to exhaust
        reduce_lr
            Whether or not to reduce the learning rate as iterations progress.
            Halves the learning rate when the training loss does not move. This
            uses torch.optim.lr_scheduler.ReduceLROnPlateau and uses the
            parameters `patience` and `threshold`
        initial_lr
            The initial learning rate to use.
        alpha
            A constant for the regularization term.
        loss_fn
            The loss to optimise for. This must accept three parameters:
            x1 (predicted), x2 (labels) and a weight vector
        reg_term
            Regularization is defined by the `reg_term` norm of the weights.
            Please use `None` if you do not wish to use regularization.
        patience
            Defines the number of iterations in a row the loss must remain
            within `threshold` in order to be classified as converged.
        threshold
            Threshold for convergence detection.
        running_loss_window
            Used to report the training loss once we have finished training and
            to determine when we have converged (along with reducing the
            learning rate).

            The reported training loss will take the last `running_loss_window`
            iterations and average them.

            If `None` we will approximate this to be the number of examples in
            an epoch.
        init_scheme
            Initialization to use prior to training the linear model.
        device
            The device to send the model and data to. If None then no `.to` call
            will be used.
        debug
            Whether to print the loss, learning rate per iteration

    Returns
        This will return the final training loss (averaged with
        `running_loss_window`)
    """
    loss_window: List[torch.Tensor] = []
    min_avg_loss = None
    convergence_counter = 0
    converged = False

    def get_point(datapoint):
        if len(datapoint) == 2:
            x, y = datapoint
            w = None
        else:
            x, y, w = datapoint

        if device is not None:
            x = x.to(device)
            y = y.to(device)
            if w is not None:
                w = w.to(device)

        return x, y, w

    # get a point and construct the model
    data_iter = iter(dataloader)
    x, y, w = get_point(next(data_iter))

    model._construct_model_params(
        in_features=x.shape[1],
        out_features=y.shape[1] if len(y.shape) == 2 else 1,
        **construct_kwargs,
    )
    model.train()

    assert model.linear is not None

    if init_scheme is not None:
        assert init_scheme in ["xavier", "zeros"]

        with torch.no_grad():
            if init_scheme == "xavier":
                torch.nn.init.xavier_uniform_(model.linear.weight)
            else:
                model.linear.weight.zero_()

            if model.linear.bias is not None:
                model.linear.bias.zero_()

    with torch.enable_grad():
        optim = torch.optim.SGD(model.parameters(), lr=initial_lr)
        if reduce_lr:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optim, factor=0.5, patience=patience, threshold=threshold
            )

        t1 = time.time()
        epoch = 0
        i = 0
        while epoch < max_epoch:
            while True:  # for x, y, w in dataloader
                if running_loss_window is None:
                    running_loss_window = x.shape[0] * len(dataloader)

                y = y.view(x.shape[0], -1)
                if w is not None:
                    w = w.view(x.shape[0], -1)

                i += 1

                out = model(x)

                loss = loss_fn(y, out, w)
                if reg_term is not None:
                    reg = torch.norm(model.linear.weight, p=reg_term)
                    loss += reg.sum() * alpha

                if len(loss_window) >= running_loss_window:
                    loss_window = loss_window[1:]
                loss_window.append(loss.clone().detach())
                assert len(loss_window) <= running_loss_window

                average_loss = torch.mean(torch.stack(loss_window))
                if min_avg_loss is not None:
                    # if we haven't improved by at least `threshold`
                    if average_loss > min_avg_loss or torch.isclose(
                        min_avg_loss, average_loss, atol=threshold
                    ):
                        convergence_counter += 1
                        if convergence_counter >= patience:
                            converged = True
                            break
                    else:
                        convergence_counter = 0
                if min_avg_loss is None or min_avg_loss >= average_loss:
                    min_avg_loss = average_loss.clone()

                if debug:
                    print(
                        f"lr={optim.param_groups[0]['lr']}, Loss={loss},"
                        + "Aloss={average_loss}, min_avg_loss={min_avg_loss}"
                    )

                loss.backward()
                optim.step()
                model.zero_grad()
                if scheduler:
                    scheduler.step(average_loss)

                temp = next(data_iter, None)
                if temp is None:
                    break
                x, y, w = get_point(temp)

            if converged:
                break

            epoch += 1
            data_iter = iter(dataloader)
            x, y, w = get_point(next(data_iter))

    t2 = time.time()
    return {
        "train_time": t2 - t1,
        "train_loss": torch.mean(torch.stack(loss_window)).item(),
        "train_iter": i,
        "train_epoch": epoch,
    }


class NormLayer(nn.Module):
    def __init__(self, mean, std, n=None, eps=1e-8) -> None:
        super().__init__()
        self.mean = mean
        self.std = std
        self.eps = eps

    def forward(self, x):
        return (x - self.mean) / (self.std + self.eps)


def sklearn_train_linear_model(
    model: LinearModel,
    dataloader: DataLoader,
    construct_kwargs: Dict[str, Any],
    sklearn_trainer: str = "Lasso",
    norm_input: bool = False,
    **fit_kwargs,
):
    r"""
    Alternative method to train with sklearn. This does introduce some slight
    overhead as we convert the tensors to numpy and then convert the resulting
    trained model to a `LinearModel` object. However, this conversion
    should be negligible.

    Please note that this assumes:

    0. You have sklearn and numpy installed
    1. The dataset can fit into memory

    Args
        model
            The model to train.
        dataloader
            The data to use. This will be exhausted and converted to numpy
            arrays. Therefore please do not feed an infinite dataloader.
        norm_input
            Whether or not to normalize the input
        sklearn_trainer
            The sklearn model to use to train the model. Please refer to
            sklearn.linear_model for a list of modules to use.
        construct_kwargs
            Additional arguments provided to the `sklearn_trainer` constructor
        fit_kwargs
            Other arguments to send to `sklearn_trainer`'s `.fit` method
    """
    from functools import reduce

    try:
        import numpy as np
    except ImportError:
        raise ValueError("numpy is not available. Please install numpy.")

    try:
        import sklearn
        import sklearn.linear_model
        import sklearn.svm
    except ImportError:
        raise ValueError("sklearn is not available. Please install sklearn >= 0.23")

    if not sklearn.__version__ >= "0.23.0":
        warnings.warn(
            "Must have sklearn version 0.23.0 or higher to use "
            "sample_weight in Lasso regression."
        )

    num_batches = 0
    xs, ys, ws = [], [], []
    for data in dataloader:
        if len(data) == 3:
            x, y, w = data
        else:
            assert len(data) == 2
            x, y = data
            w = None

        xs.append(x.cpu().numpy())
        ys.append(y.cpu().numpy())
        if w is not None:
            ws.append(w.cpu().numpy())
        num_batches += 1

    x = np.concatenate(xs, axis=0)
    y = np.concatenate(ys, axis=0)
    if len(ws) > 0:
        w = np.concatenate(ws, axis=0)
    else:
        w = None

    if norm_input:
        mean, std = x.mean(0), x.std(0)
        x -= mean
        x /= std

    t1 = time.time()
    sklearn_model = reduce(
        lambda val, el: getattr(val, el), [sklearn] + sklearn_trainer.split(".")
    )(**construct_kwargs)
    try:
        sklearn_model.fit(x, y, sample_weight=w, **fit_kwargs)
    except TypeError:
        sklearn_model.fit(x, y, **fit_kwargs)
        warnings.warn(
            "Sample weight is not supported for the provided linear model!"
            " Trained model without weighting inputs. For Lasso, please"
            " upgrade sklearn to a version >= 0.23.0."
        )

    t2 = time.time()

    # Convert weights to pytorch
    classes = (
        torch.IntTensor(sklearn_model.classes_)
        if hasattr(sklearn_model, "classes_")
        else None
    )

    # extract model device
    device = getattr(model, "device", "cpu")

    num_outputs = sklearn_model.coef_.shape[0] if sklearn_model.coef_.ndim > 1 else 1
    weight_values = torch.FloatTensor(sklearn_model.coef_).to(device)  # type: ignore
    bias_values = torch.FloatTensor([sklearn_model.intercept_]).to(  # type: ignore
        device  # type: ignore
    )  # type: ignore
    model._construct_model_params(
        norm_type=None,
        weight_values=weight_values.view(num_outputs, -1),
        bias_value=bias_values.squeeze().unsqueeze(0),
        classes=classes,
    )

    if norm_input:
        model.norm = NormLayer(mean, std)

    return {"train_time": t2 - t1}



#!/usr/bin/env python3

from __future__ import print_function

import unittest
from typing import Any, Tuple, Union

import torch
from captum._utils.typing import TensorOrTupleOfTensorsGeneric
from captum.attr._core.guided_backprop_deconvnet import Deconvolution
from captum.attr._core.neuron.neuron_guided_backprop_deconvnet import (
    NeuronDeconvolution,
)
from tests.helpers.basic import assertTensorAlmostEqual, BaseTest
from tests.helpers.basic_models import BasicModel_ConvNet_One_Conv
from torch.nn import Module


class Test(BaseTest):
    def test_simple_input_conv_deconv(self) -> None:
        net = BasicModel_ConvNet_One_Conv()
        inp = 1.0 * torch.arange(16, dtype=torch.float).view(1, 1, 4, 4)
        exp = [
            [2.0, 3.0, 3.0, 1.0],
            [3.0, 5.0, 5.0, 2.0],
            [3.0, 5.0, 5.0, 2.0],
            [1.0, 2.0, 2.0, 1.0],
        ]
        exp = torch.tensor(exp).view(1, 1, 4, 4)
        self._deconv_test_assert(net, (inp,), (exp,))

    def test_simple_input_conv_neuron_deconv(self) -> None:
        net = BasicModel_ConvNet_One_Conv()
        inp = 1.0 * torch.arange(16, dtype=torch.float).view(1, 1, 4, 4)
        exp = [
            [2.0, 3.0, 3.0, 1.0],
            [3.0, 5.0, 5.0, 2.0],
            [3.0, 5.0, 5.0, 2.0],
            [1.0, 2.0, 2.0, 1.0],
        ]
        exp = torch.tensor(exp).view(1, 1, 4, 4)
        self._neuron_deconv_test_assert(net, net.fc1, (0,), (inp,), (exp,))

    def test_simple_input_conv_neuron_deconv_agg_neurons(self) -> None:
        net = BasicModel_ConvNet_One_Conv()
        inp = 1.0 * torch.arange(16, dtype=torch.float).view(1, 1, 4, 4)
        exp = [
            [2.0, 3.0, 3.0, 1.0],
            [3.0, 5.0, 5.0, 2.0],
            [3.0, 5.0, 5.0, 2.0],
            [1.0, 2.0, 2.0, 1.0],
        ]
        exp = torch.tensor(exp).view(1, 1, 4, 4)
        self._neuron_deconv_test_assert(net, net.fc1, (slice(0, 1, 1),), (inp,), (exp,))

    def test_simple_multi_input_conv_deconv(self) -> None:
        net = BasicModel_ConvNet_One_Conv()
        inp = torch.arange(16, dtype=torch.float).view(1, 1, 4, 4)
        inp2 = torch.ones((1, 1, 4, 4))
        ex_attr = [
            [2.0, 3.0, 3.0, 1.0],
            [3.0, 5.0, 5.0, 2.0],
            [3.0, 5.0, 5.0, 2.0],
            [1.0, 2.0, 2.0, 1.0],
        ]
        ex_attr = torch.tensor(ex_attr).view(1, 1, 4, 4)
        self._deconv_test_assert(net, (inp, inp2), (ex_attr, ex_attr))

    def test_simple_multi_input_conv_neuron_deconv(self) -> None:
        net = BasicModel_ConvNet_One_Conv()
        inp = torch.arange(16, dtype=torch.float).view(1, 1, 4, 4)
        inp2 = torch.ones((1, 1, 4, 4))
        ex_attr = [
            [2.0, 3.0, 3.0, 1.0],
            [3.0, 5.0, 5.0, 2.0],
            [3.0, 5.0, 5.0, 2.0],
            [1.0, 2.0, 2.0, 1.0],
        ]
        ex_attr = torch.tensor(ex_attr).view(1, 1, 4, 4)
        self._neuron_deconv_test_assert(
            net, net.fc1, (3,), (inp, inp2), (ex_attr, ex_attr)
        )

    def test_deconv_matching(self) -> None:
        net = BasicModel_ConvNet_One_Conv()
        inp = 100.0 * torch.randn(1, 1, 4, 4)
        self._deconv_matching_assert(net, net.relu2, inp)

    def _deconv_test_assert(
        self,
        model: Module,
        test_input: TensorOrTupleOfTensorsGeneric,
        expected: Tuple[torch.Tensor, ...],
        additional_input: Any = None,
    ) -> None:
        deconv = Deconvolution(model)
        attributions = deconv.attribute(
            test_input, target=0, additional_forward_args=additional_input
        )
        for i in range(len(test_input)):
            assertTensorAlmostEqual(self, attributions[i], expected[i], delta=0.01)

    def _neuron_deconv_test_assert(
        self,
        model: Module,
        layer: Module,
        neuron_selector: Union[int, Tuple[Union[int, slice], ...]],
        test_input: TensorOrTupleOfTensorsGeneric,
        expected: Tuple[torch.Tensor, ...],
        additional_input: Any = None,
    ) -> None:
        deconv = NeuronDeconvolution(model, layer)
        attributions = deconv.attribute(
            test_input,
            neuron_selector=neuron_selector,
            additional_forward_args=additional_input,
        )
        for i in range(len(test_input)):
            assertTensorAlmostEqual(self, attributions[i], expected[i], delta=0.01)

    def _deconv_matching_assert(
        self,
        model: Module,
        output_layer: Module,
        test_input: TensorOrTupleOfTensorsGeneric,
    ) -> None:
        out = model(test_input)
        attrib = Deconvolution(model)
        self.assertFalse(attrib.multiplies_by_inputs)
        neuron_attrib = NeuronDeconvolution(model, output_layer)
        for i in range(out.shape[1]):
            deconv_vals = attrib.attribute(test_input, target=i)
            neuron_deconv_vals = neuron_attrib.attribute(test_input, (i,))
            assertTensorAlmostEqual(self, deconv_vals, neuron_deconv_vals, delta=0.01)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
from collections import defaultdict
from copy import copy
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import torch
from captum._utils.common import (
    _format_baseline,
    _format_feature_mask,
    _format_output,
    _format_tensor_into_tuples,
    _get_max_feature_index,
    _run_forward,
)
from captum._utils.typing import BaselineType
from captum.attr._core.feature_ablation import FeatureAblation
from captum.attr._utils.attribution import Attribution
from torch import Tensor


class InputRole:
    need_attr = 0
    need_forward = 1
    no_forward = 2


SUPPORTED_METHODS = {FeatureAblation}


# default reducer wehn reduce is None. Simply concat the outputs by the batch dimension
def _concat_tensors(accum, cur_output, _):
    return cur_output if accum is None else torch.cat([accum, cur_output])


def _create_perturbation_mask(
    perturbed_feature_indices: Tensor,  # 1D tensor of one-hot feature indices
    feature_mask: Tuple[Tensor, ...],
    feature_idx_to_mask_idx: Dict[int, List[int]],
) -> Tuple[Union[Tensor, None], ...]:
    """
    Create binary mask for inputs based on perturbed one-hot feature indices
    Use None if no perturbation is needed for the corresponding input
    """

    # a set of input/mask indices that need perturbation
    perturbation_mask_indices = set()
    for i, v in enumerate(perturbed_feature_indices.tolist()):
        # value 0 means the feature has been perturbed
        if not v:
            perturbation_mask_indices |= set(feature_idx_to_mask_idx[i])

    # create binary mask for inputs & set it to None if no perturbation is needed
    perturbation_mask = tuple(
        perturbed_feature_indices[mask_elem] if i in perturbation_mask_indices else None
        for i, mask_elem in enumerate(feature_mask)
    )

    return perturbation_mask


def _perturb_inputs(
    inputs: Iterable[Any],
    input_roles: Tuple[int],
    baselines: Tuple[Union[int, float, Tensor], ...],
    perturbation_mask: Tuple[Union[Tensor, None], ...],
) -> Tuple[Any, ...]:
    """
    Perturb inputs based on perturbation mask and baselines
    """

    perturbed_inputs = []
    attr_inp_count = 0

    for inp, role in zip(inputs, input_roles):
        if role != InputRole.need_attr:
            perturbed_inputs.append(inp)
            continue

        pert_mask = perturbation_mask[attr_inp_count]

        # no perturbation is needed for this input
        if pert_mask is None:
            perturbed_inputs.append(inp)
        else:
            baseline = baselines[attr_inp_count]

            perturbed_inp = inp * pert_mask + baseline * (1 - pert_mask)
            perturbed_inputs.append(perturbed_inp)

        attr_inp_count += 1

    perturbed_inputs = tuple(perturbed_inputs)

    return perturbed_inputs


def _convert_output_shape(
    unique_attr: Tensor,
    attr_inputs: Tuple[Tensor, ...],
    feature_mask: Tuple[Tensor, ...],
) -> Tuple[Tensor, ...]:
    """
    Convert the shape of a single tensor of unique feature attributionto
    to match the shape of the inputs returned by dataloader
    """

    # unique_attr in shape(*output_dims, n_features)
    output_dims = unique_attr.shape[:-1]
    n_features = unique_attr.shape[-1]

    attr = []

    for inp, mask in zip(attr_inputs, feature_mask):
        # input in shape(batch_size, *inp_feature_dims)
        # attribute in shape(*output_dims, *inp_feature_dims)
        attr_shape = (*output_dims, *inp.shape[1:])

        expanded_feature_indices = mask.expand(attr_shape)

        if len(inp.shape) > 2:
            # exclude batch_size & last of actual value
            extra_inp_dims = list(inp.shape[1:-1])

            # unsqueeze unqiue_attr to have same number of dims as inp
            # (*output_dims, 1..., 1, n_features)
            # then broadcast to (*output_dims, *inp.shape[1:-1], n_features)
            n_extra_dims = len(extra_inp_dims)
            unsqueezed_shape = (*output_dims, *(1,) * n_extra_dims, n_features)
            expanded_shape = (*output_dims, *extra_inp_dims, n_features)
            expanded_unqiue_attr = unique_attr.reshape(unsqueezed_shape).expand(
                expanded_shape
            )
        else:
            expanded_unqiue_attr = unique_attr

        # gather from (*output_dims, *inp.shape[1:-1], n_features)
        inp_attr = torch.gather(expanded_unqiue_attr, -1, expanded_feature_indices)
        attr.append(inp_attr)

    return tuple(attr)


class DataLoaderAttribution(Attribution):
    r"""
    Decorate a perturbation-based attribution algorthm to make it work with dataloaders.
    The decorated instance will calculate attribution in the
    same way as configured in the original attribution instance, but it will provide a
    new "attribute" function which accepts a pytorch "dataloader" instance as the input
    instead of a single batched "tensor" and supports customizing a "reduce" function to
    determine how the forward return of each iteration of the dataloader should be
    aggregated to single metric tensor to attribute. This would
    be specially useful to attribute against some corpus-wise metrics,
    e.g., Precision & Recall.
    """

    def __init__(self, attr_method: Attribution) -> None:
        r"""
        Args:
            attr_method (Attribution): An instance of any attribution algorithm
                        of type `Attribution`. E.g. Integrated Gradients,
                        Conductance or Saliency.
        """

        assert (
            type(attr_method) in SUPPORTED_METHODS
        ), f"DataloaderAttribution does not support {type(attr_method)}"

        super().__init__(attr_method.forward_func)

        # shallow copy is enough to avoid modifying original instance
        self.attr_method = copy(attr_method)

        self.attr_method.forward_func = self._forward_with_dataloader

    def _forward_with_dataloader(
        self,
        batched_perturbed_feature_indices: Tensor,
        dataloader: torch.utils.data.DataLoader,
        input_roles: Tuple[int],
        baselines: Tuple[Union[int, float, Tensor], ...],
        feature_mask: Tuple[Tensor, ...],
        reduce: Callable,
        to_metric: Optional[Callable],
        show_progress: bool,
        feature_idx_to_mask_idx: Dict[int, List[int]],
    ):
        """
        Wrapper of the original given forward_func to be used in the attribution method
        It iterates over the dataloader with the given forward_func
        """

        # batched_perturbed_feature_indices in shape(n_perturb, n_features)
        # n_perturb is not always the same as perturb_per_pass if not enough perturb
        perturbation_mask_list: List[Tuple[Union[Tensor, None], ...]] = [
            _create_perturbation_mask(
                perturbed_feature_indices,
                feature_mask,
                feature_idx_to_mask_idx,
            )
            for perturbed_feature_indices in batched_perturbed_feature_indices
        ]

        # each perturbation needs an accum state
        accum_states = [None for _ in range(len(perturbation_mask_list))]

        # tranverse the dataloader
        for inputs in dataloader:
            # for each batch read from the dataloader,
            # apply every perturbation based on perturbations_per_pass
            for i, perturbation_mask in enumerate(perturbation_mask_list):
                perturbed_inputs = _perturb_inputs(
                    inputs, input_roles, baselines, perturbation_mask
                )

                # due to explicitly defined roles
                # we can keep inputs in their original order
                # regardless of if they need attr
                # instead of using additional_forward_inputs
                forward_inputs = tuple(
                    _
                    for _, role in zip(perturbed_inputs, input_roles)
                    if role != InputRole.no_forward
                )

                output = _run_forward(
                    self.forward_func,
                    forward_inputs,
                )

                accum_states[i] = reduce(accum_states[i], output, perturbed_inputs)

        accum_results = [
            to_metric(accum) if to_metric else accum for accum in accum_states
        ]

        assert all(type(r) is Tensor for r in accum_results), (
            "Accumulated metrics for attribution must be a Tensor,"
            f"received: {next(r for r in accum_results if type(r) is not Tensor)}"
        )

        # shape(n_perturb * output_dims[0], *output_dims[1:])
        # the underneath attr method needs to support forward_func output's
        # 1st dim to grow with perturb_per_eval
        batched_accum = torch.stack(accum_results, dim=0)
        return batched_accum

    def attribute(
        self,
        dataloader: torch.utils.data.DataLoader,
        input_roles: Optional[Tuple[int, ...]] = None,
        baselines: BaselineType = None,
        feature_mask: Union[None, Tensor, Tuple[Tensor, ...]] = None,
        reduce: Optional[Callable] = None,
        to_metric: Optional[Callable] = None,
        perturbations_per_pass: int = 1,
        show_progress: bool = False,
        return_input_shape: bool = True,
    ) -> Union[Tensor, Tuple[Tensor, ...]]:
        r"""
        Args:

            dataloader (torch.Dataloader): the dataloader to attribute, which should
                        return a tuple of consistant size for every iteration
            input_roles (tuple[int, ...], optional): a tuple of integers to define the
                        role of each element returned from the dataloader. It should
                        have the same size as the return of the dataloader.
                        The available roles are:

                        0: the element is passed to forward_func and needs attribution.
                        It must be a tensor.
                        1: the element is excluded for forward_func. A typical example
                        is the label.
                        2: the element is passed to forward_func but does not need
                        attribution. Like additional_forward_args

            baselines (Union[Tensor, tuple[Tensor, ...]], optional): same as the
                        baseline in attribute. The same baseline will be
                        applied to the entire dataloader. The first dimension is
                        assumed to be batch size and it must be 1. Baselines should only
                        be specififed for the dataloader's returns that need
                        attribution (role = 0)

            feature_mask (Union[Tensor, tuple[Tensor, ...]], optional): same as the
                        feature_mask in attribute. The same feature_mask will be
                        applied to the entire dataloader. The first dimension is
                        assumed to be batch size and it must be 1. Mask should only
                        be specififed for the dataloader's returns that need
                        attribution (role = 0)
            reduce (Callable, optional): a function to accumulate the forward output of
                        each iteration of the dataloader. The function signature is:
                        ``reduce(accum, current_output, current_inputs) -> accum``,
                        where:

                        accum (Any): accumulated states, can be any type
                        current_output (Tensor): current output tensor from forward_func
                        current_inputs (tuple[Any,...]): current inputs from dataloader

            to_metric (Callable, optional): an optional function to further convert
                        accumulated results through "reduce" after tranversing the whole
                        dataloader to a single tensor of metrics to calculate
                        attribution against. The function signature is:
                        ``to_metric(accum) -> metric``, where:

                        accum (Any): accumulated state from reduce function
                        metric (Tensor): final result to be attributed, must be a Tensor

                        If None, will directly attribute w.r.t the reduced ``accum``
            perturbations_per_pass (int, optional) the number perturbations to execute
                        concurrently in each traverse of the dataloader. The number of
                        traverses needed is
                        ceil(n_perturbations / perturbations_per_pass).

                        This arguement offers control of the trade-off between memory
                        and efficiency. If the dataloader involves slow operations like
                        remote request or file I/O, multiple traversals can be
                        inefficient. On the other hand, each perturbation needs to
                        store its accumulated outputs of the reduce
                        function until the end of the data traverse.
            return_input_shape (bool, optional): if True, returns the attribution
                        following the input shapes given by the dataloader.
                        Otherwise, returns a single tensor for the attributions of
                        all the features, where the last dimension
                        is the number of features.

        Returns:
            **attributions** :
            - **attributions** (*Tensor* or *tuple[Tensor, ...]*):
                        Attribution with respect to each input feature.
                        if return_input_shape is True, attributions will be
                        the same size as the given dataloader's returns that need
                        attribution (role = 0), with each value
                        providing the attribution of the corresponding input index.
                        If a single tensor is provided as inputs, a single tensor is
                        returned. If a tuple is provided for inputs, a tuple of
                        corresponding sized tensors is returned.
                        If return_input_shape is False, a single tensor is returned
                        where each index of the last dimension represents a feature
        """
        inputs = next(iter(dataloader))
        is_inputs_tuple = True

        if type(inputs) is list:
            # support list as it is a common return type for dataloader in torch
            inputs = tuple(inputs)
        elif type(inputs) is not tuple:
            is_inputs_tuple = False
            inputs = _format_tensor_into_tuples(inputs)

        if input_roles:
            assert len(input_roles) == len(inputs), (
                "input_roles must have the same size as the return of the dataloader,",
                f"length of input_roles is {len(input_roles)} ",
                f"whereas the length of dataloader return is {len(inputs)}",
            )

            assert any(role == InputRole.need_attr for role in input_roles), (
                "input_roles must contain at least one element need attribution"
                f"({InputRole.need_attr}), received input_roles: {input_roles}"
            )
        else:
            # by default, assume every element in the dataloader needs attribution
            input_roles = tuple(InputRole.need_attr for _ in inputs)

        attr_inputs = tuple(
            inp for role, inp in zip(input_roles, inputs) if role == InputRole.need_attr
        )

        baselines = _format_baseline(baselines, attr_inputs)

        assert len(attr_inputs) == len(baselines), (
            "Baselines must have the same size as the return of the dataloader ",
            "that need attribution",
            f"length of baseline is {len(baselines)} ",
            f'whereas the length of dataloader return with role "0" is {len(inputs)}',
        )

        for i, baseline in enumerate(baselines):
            if isinstance(baseline, Tensor):
                assert baseline.size(0) == 1, (
                    "If the baseline is a tensor, "
                    "its 1st dim of baseline must be 1 so it can be broadacasted to "
                    "any batch of the dataloader:"
                    f"baselines[{i}].shape = {baseline.shape}"
                )

        feature_mask = _format_feature_mask(feature_mask, attr_inputs)

        assert len(attr_inputs) == len(feature_mask), (
            "Feature mask must have the same size as the return of the dataloader ",
            "that need attribution",
            f"length of feature_mask is {len(feature_mask)} ",
            f'whereas the length of dataloader return with role "0" is {len(inputs)}',
        )

        for i, each_mask in enumerate(feature_mask):
            assert each_mask.size(0) == 1, (
                "The 1st dim of feature_mask must be 1 so it can be broadcasted to "
                "any batch of the dataloader:"
                f"feature_mask[{i}].shape = {each_mask.shape}"
            )

        # map to retrieve masks contain a given feature index
        feature_idx_to_mask_idx = defaultdict(list)
        for i, mask in enumerate(feature_mask):
            unqiue_feature_indices = torch.unique(mask).tolist()
            for feature_idx in unqiue_feature_indices:
                feature_idx_to_mask_idx[feature_idx].append(i)

        max_feature_idx = _get_max_feature_index(feature_mask)
        n_features = max_feature_idx + 1

        if reduce is None:
            reduce = _concat_tensors

        # onehot tensor for feature indices
        feature_indices = torch.ones((1, n_features), device=attr_inputs[0].device)

        # unique_attr in shape(*output_dims, n_features)
        unique_attr = self.attr_method.attribute(
            feature_indices,
            perturbations_per_eval=perturbations_per_pass,
            additional_forward_args=(
                dataloader,
                input_roles,
                baselines,
                feature_mask,
                reduce,
                to_metric,
                show_progress,
                feature_idx_to_mask_idx,
            ),
        )

        if not return_input_shape:
            return unique_attr
        else:
            attr = _convert_output_shape(
                unique_attr,
                attr_inputs,
                feature_mask,
            )

            return _format_output(is_inputs_tuple, attr)



import os
import tempfile
from collections import OrderedDict
from typing import Callable, cast, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from captum.influence._core.tracincp import TracInCP
from parameterized import parameterized
from tests.helpers.basic import assertTensorAlmostEqual, BaseTest
from tests.influence._utils.common import (
    _wrap_model_in_dataparallel,
    BasicLinearNet,
    BinaryDataset,
    build_test_name_func,
    DataInfluenceConstructor,
)


class TestTracInXOR(BaseTest):

    # TODO: Move test setup to use setUp and tearDown method overrides.
    def _test_tracin_xor_setup(self, tmpdir: str, use_gpu: bool = False):
        net = BasicLinearNet(2, 2, 1)

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.2956, -1.4465], [-0.3890, -0.7420]]),
                ),
                ("linear1.bias", torch.Tensor([1.2924, 0.0021])),
                ("linear2.weight", torch.Tensor([[-1.2013, 0.7174]])),
                ("linear2.bias", torch.Tensor([0.5880])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "0" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.3238, -1.4899], [-0.4544, -0.7448]]),
                ),
                ("linear1.bias", torch.Tensor([1.3185, -0.0317])),
                ("linear2.weight", torch.Tensor([[-1.2342, 0.7741]])),
                ("linear2.bias", torch.Tensor([0.6234])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "1" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.3546, -1.5288], [-0.5250, -0.7591]]),
                ),
                ("linear1.bias", torch.Tensor([1.3432, -0.0684])),
                ("linear2.weight", torch.Tensor([[-1.2490, 0.8534]])),
                ("linear2.bias", torch.Tensor([0.6749])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "2" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.4022, -1.5485], [-0.5688, -0.7607]]),
                ),
                ("linear1.bias", torch.Tensor([1.3740, -0.1571])),
                ("linear2.weight", torch.Tensor([[-1.3412, 0.9013]])),
                ("linear2.bias", torch.Tensor([0.6468])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "3" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.4464, -1.5890], [-0.6348, -0.7665]]),
                ),
                ("linear1.bias", torch.Tensor([1.3791, -0.2008])),
                ("linear2.weight", torch.Tensor([[-1.3818, 0.9586]])),
                ("linear2.bias", torch.Tensor([0.6954])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "4" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.5217, -1.6242], [-0.6644, -0.7842]]),
                ),
                ("linear1.bias", torch.Tensor([1.3500, -0.2418])),
                ("linear2.weight", torch.Tensor([[-1.4304, 0.9980]])),
                ("linear2.bias", torch.Tensor([0.7567])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "5" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.5551, -1.6631], [-0.7420, -0.8025]]),
                ),
                ("linear1.bias", torch.Tensor([1.3508, -0.2618])),
                ("linear2.weight", torch.Tensor([[-1.4272, 1.0772]])),
                ("linear2.bias", torch.Tensor([0.8427])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "6" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        state = OrderedDict(
            [
                (
                    "linear1.weight",
                    torch.Tensor([[-1.5893, -1.6656], [-0.7863, -0.8369]]),
                ),
                ("linear1.bias", torch.Tensor([1.3949, -0.3215])),
                ("linear2.weight", torch.Tensor([[-1.4555, 1.1600]])),
                ("linear2.bias", torch.Tensor([0.8730])),
            ]
        )
        net.load_state_dict(state)
        net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net

        checkpoint_name = "-".join(["checkpoint", "class", "7" + ".pt"])
        torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        dataset = BinaryDataset(use_gpu)

        return net_adjusted, dataset

    parametrized_list = [
        (
            "none",
            DataInfluenceConstructor(
                TracInCP, name="TracInCP_linear1", layers=["linear1"]
            ),
            "check_idx",
            False,
        ),
        (
            "none",
            DataInfluenceConstructor(TracInCP, name="TracInCP_all_layers"),
            "check_idx",
            False,
        ),
        (
            None,
            DataInfluenceConstructor(TracInCP, name="TracInCP_all_layers"),
            "sample_wise_trick",
            False,
        ),
        (
            None,
            DataInfluenceConstructor(
                TracInCP, name="TracInCP_linear1_linear2", layers=["linear1", "linear2"]
            ),
            "sample_wise_trick",
            False,
        ),
    ]

    if torch.cuda.is_available() and torch.cuda.device_count() != 0:
        parametrized_list.extend(
            [
                (
                    "none",
                    DataInfluenceConstructor(TracInCP, name="TracInCP_all_layers"),
                    "check_idx",
                    True,
                ),
                (
                    "none",
                    DataInfluenceConstructor(
                        TracInCP,
                        name="TracInCP_linear1_linear2",
                        layers=["module.linear1", "module.linear2"],
                    ),
                    "check_idx",
                    True,
                ),
            ],
        )

    @parameterized.expand(
        parametrized_list,
        name_func=build_test_name_func(args_to_skip=["reduction"]),
    )
    def test_tracin_xor(
        self,
        reduction: Optional[str],
        tracin_constructor: Callable,
        mode: str,
        use_gpu: bool,
    ) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            batch_size = 4

            net, dataset = self._test_tracin_xor_setup(tmpdir, use_gpu)

            testset = F.normalize(torch.empty(100, 2).normal_(mean=0, std=0.5), dim=1)
            mask = ~torch.logical_xor(testset[:, 0] > 0, testset[:, 1] > 0)
            testlabels = (
                torch.where(mask, torch.tensor(1), torch.tensor(-1))
                .unsqueeze(1)
                .float()
            )
            if use_gpu:
                testset = testset.cuda()
                testlabels = testlabels.cuda()

            self.assertTrue(callable(tracin_constructor))

            if mode == "check_idx":

                self.assertTrue(isinstance(reduction, str))
                criterion = nn.MSELoss(reduction=cast(str, reduction))

                tracin = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                )
                test_scores = tracin.influence((testset, testlabels))
                idx = torch.argsort(test_scores, dim=1, descending=True)
                # check that top 5 influences have matching binary classification
                for i in range(len(idx)):
                    influence_labels = dataset.labels[idx[i][0:5], 0]
                    self.assertTrue(torch.all(testlabels[i, 0] == influence_labels))

            if mode == "sample_wise_trick":

                criterion = nn.MSELoss(reduction="none")

                tracin = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                    sample_wise_grads_per_batch=False,
                )

                # With sample-wise trick
                criterion = nn.MSELoss(reduction="sum")
                tracin_sample_wise_trick = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                    sample_wise_grads_per_batch=True,
                )
                test_scores = tracin.influence((testset, testlabels))
                test_scores_sample_wise_trick = tracin_sample_wise_trick.influence(
                    (testset, testlabels)
                )
                assertTensorAlmostEqual(
                    self, test_scores, test_scores_sample_wise_trick
                )



#!/usr/bin/env python3

from inspect import signature
from typing import Callable, List, Tuple, Union

import torch
from captum.attr._core.deep_lift import DeepLift, DeepLiftShap
from captum.attr._core.integrated_gradients import IntegratedGradients
from tests.helpers.basic import (
    assertAttributionComparision,
    assertTensorAlmostEqual,
    BaseTest,
)
from tests.helpers.basic_models import (
    BasicModelWithReusedModules,
    Conv1dSeqModel,
    LinearMaxPoolLinearModel,
    ReLUDeepLiftModel,
    ReLULinearModel,
    TanhDeepLiftModel,
)
from torch import Tensor
from torch.nn import Module


class Test(BaseTest):
    def test_relu_deeplift(self) -> None:
        x1 = torch.tensor([1.0], requires_grad=True)
        x2 = torch.tensor([2.0], requires_grad=True)

        b1 = torch.tensor([0.0], requires_grad=True)
        b2 = torch.tensor([0.0], requires_grad=True)

        inputs = (x1, x2)
        baselines = (b1, b2)

        model = ReLUDeepLiftModel()
        self._deeplift_assert(model, DeepLift(model), inputs, baselines)

    def test_relu_deeplift_exact_match(self) -> None:
        x1 = torch.tensor([1.0], requires_grad=True)
        x2 = torch.tensor([2.0], requires_grad=True)

        b1 = torch.tensor([0.0], requires_grad=True)
        b2 = torch.tensor([0.0], requires_grad=True)

        inputs = (x1, x2)
        baselines = (b1, b2)
        model = ReLUDeepLiftModel()
        dl = DeepLift(model)
        attributions, delta = dl.attribute(
            inputs, baselines, return_convergence_delta=True
        )
        self.assertEqual(attributions[0][0], 2.0)
        self.assertEqual(attributions[1][0], 1.0)
        self.assertEqual(delta[0], 0.0)

    def test_relu_deeplift_exact_match_wo_mutliplying_by_inputs(self) -> None:
        x1 = torch.tensor([1.0])
        x2 = torch.tensor([2.0])
        inputs = (x1, x2)

        model = ReLUDeepLiftModel()
        dl = DeepLift(model, multiply_by_inputs=False)
        attributions = dl.attribute(inputs)
        self.assertEqual(attributions[0][0], 2.0)
        self.assertEqual(attributions[1][0], 0.5)

    def test_tanh_deeplift(self) -> None:
        x1 = torch.tensor([-1.0], requires_grad=True)
        x2 = torch.tensor([-2.0], requires_grad=True)

        b1 = torch.tensor([0.0], requires_grad=True)
        b2 = torch.tensor([0.0], requires_grad=True)

        inputs = (x1, x2)
        baselines = (b1, b2)

        model = TanhDeepLiftModel()
        self._deeplift_assert(model, DeepLift(model), inputs, baselines)

    def test_relu_deeplift_batch(self) -> None:
        x1 = torch.tensor([[1.0], [1.0], [1.0], [1.0]], requires_grad=True)
        x2 = torch.tensor([[2.0], [2.0], [2.0], [2.0]], requires_grad=True)

        b1 = torch.tensor([[0.0], [0.0], [0.0], [0.0]], requires_grad=True)
        b2 = torch.tensor([[0.0], [0.0], [0.0], [0.0]], requires_grad=True)

        inputs = (x1, x2)
        baselines = (b1, b2)

        model = ReLUDeepLiftModel()
        self._deeplift_assert(model, DeepLift(model), inputs, baselines)

    def test_relu_linear_deeplift(self) -> None:
        model = ReLULinearModel(inplace=False)
        x1 = torch.tensor([[-10.0, 1.0, -5.0]], requires_grad=True)
        x2 = torch.tensor([[3.0, 3.0, 1.0]], requires_grad=True)

        inputs = (x1, x2)
        baselines = (0, 0.0001)

        # expected = [[[0.0, 0.0]], [[6.0, 2.0]]]
        self._deeplift_assert(model, DeepLift(model), inputs, baselines)

    def test_relu_linear_deeplift_compare_inplace(self) -> None:
        model1 = ReLULinearModel(inplace=True)
        x1 = torch.tensor([[-10.0, 1.0, -5.0], [2.0, 3.0, 4.0]], requires_grad=True)
        x2 = torch.tensor([[3.0, 3.0, 1.0], [2.3, 5.0, 4.0]], requires_grad=True)
        inputs = (x1, x2)
        attributions1 = DeepLift(model1).attribute(inputs)

        model2 = ReLULinearModel()
        attributions2 = DeepLift(model2).attribute(inputs)
        assertTensorAlmostEqual(self, attributions1[0], attributions2[0])
        assertTensorAlmostEqual(self, attributions1[1], attributions2[1])

    def test_relu_linear_deepliftshap_compare_inplace(self) -> None:
        model1 = ReLULinearModel(inplace=True)
        x1 = torch.tensor([[-10.0, 1.0, -5.0], [2.0, 3.0, 4.0]], requires_grad=True)
        x2 = torch.tensor([[3.0, 3.0, 1.0], [2.3, 5.0, 4.0]], requires_grad=True)
        inputs = (x1, x2)
        b1 = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
        b2 = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
        baselines = (b1, b2)

        attributions1 = DeepLiftShap(model1).attribute(inputs, baselines)

        model2 = ReLULinearModel()
        attributions2 = DeepLiftShap(model2).attribute(inputs, baselines)
        assertTensorAlmostEqual(self, attributions1[0], attributions2[0])
        assertTensorAlmostEqual(self, attributions1[1], attributions2[1])

    def test_relu_linear_deeplift_batch(self) -> None:
        model = ReLULinearModel(inplace=True)
        x1 = torch.tensor([[-10.0, 1.0, -5.0], [2.0, 3.0, 4.0]], requires_grad=True)
        x2 = torch.tensor([[3.0, 3.0, 1.0], [2.3, 5.0, 4.0]], requires_grad=True)

        inputs = (x1, x2)
        baselines = (torch.zeros(1, 3), torch.rand(1, 3) * 0.001)
        # expected = [[[0.0, 0.0]], [[6.0, 2.0]]]
        self._deeplift_assert(model, DeepLift(model), inputs, baselines)

    def test_relu_deeplift_with_hypothetical_contrib_func(self) -> None:
        model = Conv1dSeqModel()
        rand_seq_data = torch.abs(torch.randn(2, 4, 1000))
        rand_seq_ref = torch.abs(torch.randn(2, 4, 1000))
        dls = DeepLift(model)
        attr = dls.attribute(
            rand_seq_data,
            rand_seq_ref,
            custom_attribution_func=_hypothetical_contrib_func,
            target=(1, 0),
        )
        self.assertEqual(attr.shape, rand_seq_data.shape)

    def test_relu_deepliftshap_batch_4D_input(self) -> None:
        x1 = torch.ones(4, 1, 1, 1)
        x2 = torch.tensor([[[[2.0]]]] * 4)

        b1 = torch.zeros(4, 1, 1, 1)
        b2 = torch.zeros(4, 1, 1, 1)

        inputs = (x1, x2)
        baselines = (b1, b2)

        model = ReLUDeepLiftModel()
        self._deeplift_assert(model, DeepLiftShap(model), inputs, baselines)

    def test_relu_deepliftshap_batch_4D_input_wo_mutliplying_by_inputs(self) -> None:
        x1 = torch.ones(4, 1, 1, 1)
        x2 = torch.tensor([[[[2.0]]]] * 4)

        b1 = torch.zeros(4, 1, 1, 1)
        b2 = torch.zeros(4, 1, 1, 1)

        inputs = (x1, x2)
        baselines = (b1, b2)

        model = ReLUDeepLiftModel()
        attr = DeepLiftShap(model, multiply_by_inputs=False).attribute(
            inputs, baselines
        )
        assertTensorAlmostEqual(self, attr[0], 2 * torch.ones(4, 1, 1, 1))
        assertTensorAlmostEqual(self, attr[1], 0.5 * torch.ones(4, 1, 1, 1))

    def test_relu_deepliftshap_multi_ref(self) -> None:
        x1 = torch.tensor([[1.0]], requires_grad=True)
        x2 = torch.tensor([[2.0]], requires_grad=True)

        b1 = torch.tensor([[0.0], [0.0], [0.0], [0.0]], requires_grad=True)
        b2 = torch.tensor([[0.0], [0.0], [0.0], [0.0]], requires_grad=True)

        inputs = (x1, x2)
        baselines = (b1, b2)

        model = ReLUDeepLiftModel()
        self._deeplift_assert(model, DeepLiftShap(model), inputs, baselines)

    def test_relu_deepliftshap_baselines_as_func(self) -> None:
        model = ReLULinearModel(inplace=True)
        x1 = torch.tensor([[-10.0, 1.0, -5.0]])
        x2 = torch.tensor([[3.0, 3.0, 1.0]])

        def gen_baselines() -> Tuple[Tensor, ...]:
            b1 = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
            b2 = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
            return (b1, b2)

        def gen_baselines_scalar() -> Tuple[float, ...]:
            return (0.0, 0.0001)

        def gen_baselines_with_inputs(inputs: Tuple[Tensor, ...]) -> Tuple[Tensor, ...]:
            b1 = torch.cat([inputs[0], inputs[0] - 10])
            b2 = torch.cat([inputs[1], inputs[1] - 10])
            return (b1, b2)

        def gen_baselines_returns_array() -> Tuple[List[List[float]], ...]:
            b1 = [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]]
            b2 = [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]]
            return (b1, b2)

        inputs = (x1, x2)

        dl_shap = DeepLiftShap(model)
        self._deeplift_assert(model, dl_shap, inputs, gen_baselines)
        self._deeplift_assert(model, dl_shap, inputs, gen_baselines_with_inputs)
        with self.assertRaises(AssertionError):
            self._deeplift_assert(
                model, DeepLiftShap(model), inputs, gen_baselines_returns_array
            )
        with self.assertRaises(AssertionError):
            self._deeplift_assert(model, dl_shap, inputs, gen_baselines_scalar)

        baselines = gen_baselines()
        attributions = dl_shap.attribute(inputs, baselines)
        attributions_with_func = dl_shap.attribute(inputs, gen_baselines)
        assertTensorAlmostEqual(self, attributions[0], attributions_with_func[0])
        assertTensorAlmostEqual(self, attributions[1], attributions_with_func[1])

    def test_relu_deepliftshap_with_custom_attr_func(self) -> None:
        def custom_attr_func(
            multipliers: Tuple[Tensor, ...],
            inputs: Tuple[Tensor, ...],
            baselines: Tuple[Tensor, ...],
        ) -> Tuple[Tensor, ...]:
            return tuple(multiplier * 0.0 for multiplier in multipliers)

        model = ReLULinearModel(inplace=True)
        x1 = torch.tensor([[-10.0, 1.0, -5.0]])
        x2 = torch.tensor([[3.0, 3.0, 1.0]])
        b1 = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
        b2 = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
        inputs = (x1, x2)
        baselines = (b1, b2)
        dls = DeepLiftShap(model)
        attr_w_func = dls.attribute(
            inputs, baselines, custom_attribution_func=custom_attr_func
        )

        assertTensorAlmostEqual(self, attr_w_func[0], [[0.0, 0.0, 0.0]], 0.0)
        assertTensorAlmostEqual(self, attr_w_func[1], [[0.0, 0.0, 0.0]], 0.0)

    def test_relu_deepliftshap_with_hypothetical_contrib_func(self) -> None:
        model = Conv1dSeqModel()
        rand_seq_data = torch.abs(torch.randn(2, 4, 1000))
        rand_seq_ref = torch.abs(torch.randn(3, 4, 1000))
        dls = DeepLiftShap(model)
        attr = dls.attribute(
            rand_seq_data,
            rand_seq_ref,
            custom_attribution_func=_hypothetical_contrib_func,
            target=(0, 0),
        )
        self.assertEqual(attr.shape, rand_seq_data.shape)

    def test_reusable_modules(self) -> None:
        model = BasicModelWithReusedModules()
        input = torch.rand(1, 3)
        dl = DeepLift(model)
        with self.assertRaises(RuntimeError):
            dl.attribute(input, target=0)

    def test_lin_maxpool_lin_classification(self) -> None:
        inputs = torch.ones(2, 4)
        baselines = torch.tensor([[1, 2, 3, 9], [4, 8, 6, 7]]).float()

        model = LinearMaxPoolLinearModel()
        dl = DeepLift(model)
        attrs, delta = dl.attribute(
            inputs, baselines, target=0, return_convergence_delta=True
        )
        expected = torch.Tensor([[0.0, 0.0, 0.0, -8.0], [0.0, -7.0, 0.0, 0.0]])
        expected_delta = torch.Tensor([0.0, 0.0])
        assertTensorAlmostEqual(self, attrs, expected, 0.0001)
        assertTensorAlmostEqual(self, delta, expected_delta, 0.0001)

    def _deeplift_assert(
        self,
        model: Module,
        attr_method: Union[DeepLift, DeepLiftShap],
        inputs: Tuple[Tensor, ...],
        baselines,
        custom_attr_func: Callable[..., Tuple[Tensor, ...]] = None,
    ) -> None:
        input_bsz = len(inputs[0])
        if callable(baselines):
            baseline_parameters = signature(baselines).parameters
            if len(baseline_parameters) > 0:
                baselines = baselines(inputs)
            else:
                baselines = baselines()

        baseline_bsz = (
            len(baselines[0]) if isinstance(baselines[0], torch.Tensor) else 1
        )
        # Run attribution multiple times to make sure that it is
        # working as expected
        for _ in range(5):
            model.zero_grad()
            attributions, delta = attr_method.attribute(
                inputs,
                baselines,
                return_convergence_delta=True,
                custom_attribution_func=custom_attr_func,
            )
            attributions_without_delta = attr_method.attribute(
                inputs, baselines, custom_attribution_func=custom_attr_func
            )

            for attribution, attribution_without_delta in zip(
                attributions, attributions_without_delta
            ):
                self.assertTrue(
                    torch.all(torch.eq(attribution, attribution_without_delta))
                )

            if isinstance(attr_method, DeepLiftShap):
                self.assertEqual([input_bsz * baseline_bsz], list(delta.shape))
            else:
                self.assertEqual([input_bsz], list(delta.shape))
                delta_external = attr_method.compute_convergence_delta(
                    attributions, baselines, inputs
                )
                assertTensorAlmostEqual(
                    self, delta, delta_external, delta=0.0, mode="max"
                )

            delta_condition = (delta.abs() < 0.00001).all()
            self.assertTrue(
                delta_condition,
                "The sum of attribution values {} is not "
                "nearly equal to the difference between the endpoint for "
                "some samples".format(delta),
            )
            for input, attribution in zip(inputs, attributions):
                self.assertEqual(input.shape, attribution.shape)
            if (
                isinstance(baselines[0], (int, float))
                or inputs[0].shape == baselines[0].shape
            ):
                # Compare with Integrated Gradients
                ig = IntegratedGradients(model)
                attributions_ig = ig.attribute(inputs, baselines)
                assertAttributionComparision(self, attributions, attributions_ig)


def _hypothetical_contrib_func(
    multipliers: Tuple[Tensor, ...],
    inputs: Tuple[Tensor, ...],
    baselines: Tuple[Tensor, ...],
) -> Tuple[Tensor, ...]:
    r"""
    Implements hypothetical input contributions based on the logic described here:
    https://github.com/kundajelab/deeplift/pull/36/files
    This is using a dummy model for test purposes
    """
    # we assume that multiplies, inputs and baselines have the following shape:
    # tuple((bsz x len x channel), )
    assert len(multipliers[0].shape) == 3, multipliers[0].shape
    assert len(inputs[0].shape) == 3, inputs[0].shape
    assert len(baselines[0].shape) == 3, baselines[0].shape
    assert len(multipliers) == len(inputs) and len(inputs) == len(baselines), (
        "multipliers, inputs and baselines must have the same shape but"
        "multipliers: {}, inputs: {}, baselines: {}".format(
            len(multipliers), len(inputs), len(baselines)
        )
    )

    attributions = []
    for k in range(len(multipliers)):
        sub_attributions = torch.zeros_like(inputs[k])
        for i in range(inputs[k].shape[-1]):
            hypothetical_input = torch.zeros_like(inputs[k])
            hypothetical_input[:, :, i] = 1.0
            hypothetical_input_ref_diff = hypothetical_input - baselines[k]
            sub_attributions[:, :, i] = torch.sum(
                hypothetical_input_ref_diff * multipliers[k], dim=-1
            )
        attributions.append(sub_attributions)
    return tuple(attributions)



#!/usr/bin/env python3
from typing import Any, cast, Tuple, Union

import torch
from captum._utils.gradient import compute_gradients
from captum._utils.typing import TensorOrTupleOfTensorsGeneric
from captum.attr._core.noise_tunnel import NoiseTunnel
from captum.attr._core.saliency import Saliency
from tests.helpers.basic import (
    assertTensorAlmostEqual,
    assertTensorTuplesAlmostEqual,
    BaseTest,
)
from tests.helpers.basic_models import BasicModel, BasicModel5_MultiArgs
from tests.helpers.classification_models import SoftmaxModel
from torch import Tensor
from torch.nn import Module


def _get_basic_config() -> Tuple[Module, Tensor, Tensor, Any]:
    input = torch.tensor([1.0, 2.0, 3.0, 0.0, -1.0, 7.0], requires_grad=True).T
    # manually percomputed gradients
    grads = torch.tensor([-0.0, -0.0, -0.0, 1.0, 1.0, -0.0])
    return BasicModel(), input, grads, None


def _get_multiargs_basic_config() -> Tuple[
    Module, Tuple[Tensor, ...], Tuple[Tensor, ...], Any
]:
    model = BasicModel5_MultiArgs()
    additional_forward_args = ([2, 3], 1)
    inputs = (
        torch.tensor([[1.5, 2.0, 34.3], [3.4, 1.2, 2.0]], requires_grad=True),
        torch.tensor([[3.0, 3.5, 23.2], [2.3, 1.2, 0.3]], requires_grad=True),
    )
    grads = compute_gradients(
        model, inputs, additional_forward_args=additional_forward_args
    )
    return model, inputs, grads, additional_forward_args


def _get_multiargs_basic_config_large() -> Tuple[
    Module, Tuple[Tensor, ...], Tuple[Tensor, ...], Any
]:
    model = BasicModel5_MultiArgs()
    additional_forward_args = ([2, 3], 1)
    inputs = (
        torch.tensor(
            [[10.5, 12.0, 34.3], [43.4, 51.2, 32.0]], requires_grad=True
        ).repeat_interleave(3, dim=0),
        torch.tensor(
            [[1.0, 3.5, 23.2], [2.3, 1.2, 0.3]], requires_grad=True
        ).repeat_interleave(3, dim=0),
    )
    grads = compute_gradients(
        model, inputs, additional_forward_args=additional_forward_args
    )
    return model, inputs, grads, additional_forward_args


class Test(BaseTest):
    def test_saliency_test_basic_vanilla(self) -> None:
        self._saliency_base_assert(*_get_basic_config())

    def test_saliency_test_basic_smoothgrad(self) -> None:
        self._saliency_base_assert(*_get_basic_config(), nt_type="smoothgrad")

    def test_saliency_test_basic_vargrad(self) -> None:
        self._saliency_base_assert(*_get_basic_config(), nt_type="vargrad")

    def test_saliency_test_basic_multi_variable_vanilla(self) -> None:
        self._saliency_base_assert(*_get_multiargs_basic_config())

    def test_saliency_test_basic_multi_variable_smoothgrad(self) -> None:
        self._saliency_base_assert(*_get_multiargs_basic_config(), nt_type="smoothgrad")

    def test_saliency_test_basic_multivar_sg_n_samples_batch_size_2(self) -> None:
        attributions_batch_size = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="smoothgrad",
            n_samples_batch_size=2,
        )
        attributions = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="smoothgrad",
        )

        assertTensorTuplesAlmostEqual(self, attributions_batch_size, attributions)

    def test_saliency_test_basic_multivar_sg_n_samples_batch_size_3(self) -> None:
        attributions_batch_size = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="smoothgrad_sq",
            n_samples_batch_size=3,
        )
        attributions = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="smoothgrad_sq",
        )
        assertTensorTuplesAlmostEqual(self, attributions_batch_size, attributions)

    def test_saliency_test_basic_multivar_vg_n_samples_batch_size_1(self) -> None:
        attributions_batch_size = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="vargrad",
            n_samples_batch_size=1,
        )
        attributions = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="vargrad",
        )
        assertTensorTuplesAlmostEqual(self, attributions_batch_size, attributions)

    def test_saliency_test_basic_multivar_vg_n_samples_batch_size_6(self) -> None:
        attributions_batch_size = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="vargrad",
            n_samples_batch_size=6,
        )
        attributions = self._saliency_base_assert(
            *_get_multiargs_basic_config_large(),
            nt_type="vargrad",
        )
        assertTensorTuplesAlmostEqual(self, attributions_batch_size, attributions)

    def test_saliency_test_basic_multi_vargrad(self) -> None:
        self._saliency_base_assert(*_get_multiargs_basic_config(), nt_type="vargrad")

    def test_saliency_classification_vanilla(self) -> None:
        self._saliency_classification_assert()

    def test_saliency_classification_smoothgrad(self) -> None:
        self._saliency_classification_assert(nt_type="smoothgrad")

    def test_saliency_classification_vargrad(self) -> None:
        self._saliency_classification_assert(nt_type="vargrad")

    def test_saliency_grad_unchanged(self) -> None:
        model, inp, grads, add_args = _get_basic_config()
        inp.grad = torch.randn_like(inp)
        grad = inp.grad.detach().clone()
        self._saliency_base_assert(model, inp, grads, add_args)
        assertTensorTuplesAlmostEqual(self, inp.grad, grad, delta=0.0)

    def _saliency_base_assert(
        self,
        model: Module,
        inputs: TensorOrTupleOfTensorsGeneric,
        expected: TensorOrTupleOfTensorsGeneric,
        additional_forward_args: Any = None,
        nt_type: str = "vanilla",
        n_samples_batch_size=None,
    ) -> Union[Tensor, Tuple[Tensor, ...]]:
        saliency = Saliency(model)

        self.assertFalse(saliency.multiplies_by_inputs)

        if nt_type == "vanilla":
            attributions = saliency.attribute(
                inputs, additional_forward_args=additional_forward_args
            )
        else:
            nt = NoiseTunnel(saliency)
            attributions = nt.attribute(
                inputs,
                nt_type=nt_type,
                nt_samples=10,
                nt_samples_batch_size=n_samples_batch_size,
                stdevs=0.0000002,
                additional_forward_args=additional_forward_args,
            )

        for input, attribution, expected_attr in zip(inputs, attributions, expected):
            if nt_type == "vanilla":
                self._assert_attribution(attribution, expected_attr)
            self.assertEqual(input.shape, attribution.shape)

        return attributions

    def _assert_attribution(self, attribution: Tensor, expected: Tensor) -> None:
        expected = torch.abs(expected)
        if len(attribution.shape) == 0:
            assert (attribution - expected).abs() < 0.001
        else:
            assertTensorAlmostEqual(self, expected, attribution, delta=0.5, mode="max")

    def _saliency_classification_assert(self, nt_type: str = "vanilla") -> None:
        num_in = 5
        input = torch.tensor([[0.0, 1.0, 2.0, 3.0, 4.0]], requires_grad=True)
        target = torch.tensor(5)
        # 10-class classification model
        model = SoftmaxModel(num_in, 20, 10)
        saliency = Saliency(model)

        if nt_type == "vanilla":
            attributions = saliency.attribute(input, target)

            output = model(input)[:, target]
            output.backward()
            expected = torch.abs(cast(Tensor, input.grad))
            assertTensorAlmostEqual(self, attributions, expected)
        else:
            nt = NoiseTunnel(saliency)
            attributions = nt.attribute(
                input, nt_type=nt_type, nt_samples=10, stdevs=0.0002, target=target
            )
        self.assertEqual(input.shape, attributions.shape)
